import pytest
import pdb
import inspect

import logging

import autopilot
from autopilot.utils import registry
from autopilot.core.loggers import init_logger
from autopilot.hardware import Hardware
from autopilot.hardware.gpio import GPIO

_EXPECTED_HARDWARE = (
 "autopilot.hardware.cameras.Camera",
 "autopilot.hardware.cameras.Camera_CV",
 "autopilot.hardware.cameras.Camera_Spinnaker",
 "autopilot.hardware.gpio.Digital_In",
 "autopilot.hardware.gpio.Digital_Out",
 "autopilot.hardware.gpio.GPIO",
 "autopilot.hardware.gpio.LED_RGB",
 "autopilot.hardware.gpio.PWM",
 "autopilot.hardware.gpio.Solenoid",
 "autopilot.hardware.i2c.I2C_9DOF",
 "autopilot.hardware.i2c.MLX90640",
 "autopilot.hardware.usb.Scale",
 "autopilot.hardware.usb.Wheel"
)
"""A list of all the hardware we expect to have at the moment.

This doesn't need to be maintained *exactly*, but is just used as an
independent source of expectation for which Hardware objects we can expect.

So in all tests that use it, this tests a **minimal** expectation, ie. that we get all
the values that we should get if this were up to date, knowing that it might not be."""

@pytest.fixture
def logger_registry_get(caplog):
    logger = init_logger(module_name='registry', class_name='get')
    caplog.set_level(logging.DEBUG, logger='registry.get')
    return logger


@pytest.mark.parametrize(
    "base_class,class_name",
    [
        ('hardware', 'PWM'),
        ('task', 'Nafc'),
        (registry.REGISTRIES.HARDWARE, 'LED_RGB'),
        (registry.REGISTRIES.TASK, 'Free_Water'),
        (Hardware, 'Digital_In')
    ])
def test_get_one(base_class, class_name):
    """
    Get one autopilot object with a specified base class and class name using a string,
    an enum in :ref:`autopilot.utils.registry.REGISTRIES`, or an object itself
    """
    got_class = autopilot.get(base_class, class_name)
    assert inspect.isclass(got_class)
    assert got_class.__name__ == class_name

@pytest.mark.parametrize("base_class",['hardware','task',*registry.REGISTRIES])
def test_get_all(base_class):
    """
    Test that calling ``get`` with no ``class_name`` argument returns all the objects for that registry
    """
    # test that we get them
    got_classes = autopilot.get(base_class)
    assert isinstance(got_classes, list)
    assert len(got_classes)>1

    # then for the manually specified ones, test if we have the minimum classes we expect
    if isinstance(base_class, str) and base_class == 'hardware':
        class_names = ['.'.join([cls.__module__, cls.__name__]) for cls in got_classes]
        assert all([expected in class_names for expected in _EXPECTED_HARDWARE])
        assert all([issubclass(cls, Hardware) or cls is Hardware for cls in got_classes])\

def test_get_subtree(logger_registry_get, caplog):
    """
    Test that calling ``get`` with a child of a top-level object (eg ``GPIO`` rather than ``Hardware``)
    gets all its children, (using GPIO as the test case)
    """
    minimum_expected = [cls for cls in _EXPECTED_HARDWARE if 'hardware.gpio' in cls]

    # assert we haven't gotten any warnings yet
    assert all([record.levelname != "WARNING" for record in caplog.records])
    gpio_subclasses = autopilot.get('autopilot.hardware.gpio.GPIO', include_base=True)
    # we should be warned about doing this because it's in a very generic 'else' clause for now
    assert sum([record.levelname == "WARNING" for record in caplog.records]) == 1
    assert any(['Attempting to get subclasses' in record.getMessage() for record in caplog.records])

    # then test that we actually got them
    got_names = ['.'.join([cls.__module__, cls.__name__]) for cls in gpio_subclasses]

    assert all([testcls in got_names for testcls in minimum_expected])
    assert all([('hardware.gpio' in clsname) or ('autopilot.plugins' in clsname) for clsname in got_names])
    for got_cls in gpio_subclasses:
        assert issubclass(got_cls, GPIO) or got_cls is GPIO
        



def test_get_hardware():
    """
    use the :func:`autopilot.utils.registry.get_hardware` alias

    mostly a formality to keep it working since the underlying function is tested elsewhere
    """
    got_class = autopilot.get_hardware('GPIO')
    assert got_class is GPIO

def test_get_task():
    """
    use the :func:`autopilot.utils.registry.get_task` alias

    mostly a formality to keep it working since the underlying function is tested elsewhere
    """
    got_class = autopilot.get_task('Nafc')
    assert inspect.isclass(got_class)
    assert got_class.__name__ == "Nafc"

def test_get_equivalence():
    """
    Test that the same object is gotten regardless of method of specifying base_class
    """
    with_object = autopilot.get(Hardware, 'Digital_In')
    with_str = autopilot.get('hardware', 'Digital_In')
    with_enum = autopilot.get(registry.REGISTRIES.HARDWARE, 'Digital_In')

    assert with_object is with_str
    assert with_str is with_enum

def test_except_on_failure():
    """
    Ensure a exceptions are raised for nonsense
    """
    with pytest.raises(ImportError):
        autopilot.get('plausible.but.fake.module')

    with pytest.raises(ValueError):
        autopilot.get('IMPLAUSIBLE AND FAKE MODULEGgasregiluh')

    with pytest.raises(ValueError):
        autopilot.get('hardware', 'Real_Base_Class_But_Fake_But_Plausible_Class_Name')

    with pytest.raises(ValueError):
        autopilot.get('task', 'Real_Base_Class_But implwaug0219048iuhl')






"""
Sub-tasks that serve as children to other tasks.

.. note::

    The Child agent will be formalized in an upcoming release, until then these classes
    remain relatively undocumented as their design will likely change.

"""

from collections import OrderedDict as odict
from collections import deque

import autopilot.transform
from autopilot import prefs
from autopilot.hardware.gpio import Digital_Out
from autopilot.hardware.usb import Wheel
from autopilot.hardware import cameras
from autopilot.networking import Net_Node
from autopilot.core.loggers import init_logger
from autopilot.transform import transforms
from autopilot.hardware.i2c import I2C_9DOF
from autopilot.hardware.cameras import PiCamera
from autopilot.tasks import Task
from itertools import cycle
from queue import Empty, LifoQueue
import threading
import logging
from time import sleep

class Child(object):
    """Just a placeholder class for now to work with :func:`autopilot.get`"""

class Wheel_Child(Child):
    STAGE_NAMES = ['collect']

    PARAMS = odict()
    PARAMS['fs'] = {'tag': 'Velocity Reporting Rate (Hz)',
                    'type': 'int'}
    PARAMS['thresh'] = {'tag': 'Distance Threshold',
                        'type': 'int'}

    HARDWARE = {
        "OUTPUT": Digital_Out,
        "WHEEL":  Wheel
    }



    def __init__(self, stage_block=None, fs=10, thresh=100, **kwargs):
        super(Wheel_Child, self).__init__(**kwargs)
        self.fs = fs
        self.thresh = thresh

        self.hardware = {}
        self.hardware['OUTPUT'] = Digital_Out(prefs.get('HARDWARE')['OUTPUT'])
        self.hardware['WHEEL'] = Wheel(digi_out = self.hardware['OUTPUT'],
                                       fs       = self.fs,
                                       thresh   = self.thresh,
                                       mode     = "steady")
        self.stages = cycle([self.noop])
        self.stage_block = stage_block

    def noop(self):
        # just fitting in with the task structure.
        self.stage_block.clear()
        return {}

    def end(self):
        self.hardware['WHEEL'].release()
        self.stage_block.set()


class Video_Child(Child):
    PARAMS = odict()
    PARAMS['cams'] = {'tag': 'Dictionary of camera params, or list of dicts',
                      'type': ('dict', 'list')}

    def __init__(self, cams=None, stage_block = None, start_now=True, **kwargs):
        """
        Args:
            cams (dict, list): Should be a dictionary of camera parameters or a list of dicts. Dicts should have, at least::

                {
                    'type': 'string_of_camera_class',
                    'name': 'name_of_camera_in_task',
                    'param1': 'first_param'
                }
        """
        super(Video_Child, self).__init__(**kwargs)

        if cams is None:
            Exception('Need to give us a cams dictionary!')

        self.cams = {}

        self.start_now = start_now


        if isinstance(cams, dict):

            try:
                cam_class = getattr(cameras, cams['type'])
                self.cams[cams['name']] = cam_class(**cams)
                # if start:
                #     self.cams[cams['name']].capture()
            except AttributeError:
                AttributeError("Camera type {} not found!".format(cams['type']))

        elif isinstance(cams, list):
            for cam in cams:
                try:
                    cam_class = getattr(cameras, cam['type'])
                    self.cams[cam['name']] = cam_class(**cam)
                    # if start:
                    #     self.cams[cam['name']].capture()
                except AttributeError:
                    AttributeError("Camera type {} not found!".format(cam['type']))

        self.stages = cycle([self.noop])
        self.stage_block = stage_block


        if self.start_now:
            self.start()
        # self.thread = threading.Thread(target=self._stream)
        # self.thread.daemon = True
        # self.thread.start()

    def start(self):
        for cam in self.cams.values():
            cam.capture()

    def stop(self):
        for cam_name, cam in self.cams.items():
            try:
                cam.release()
            except Exception as e:
                Warning('Couldnt release camera {},\n{}'.format(cam_name, e))



    def _stream(self):
        self.node = Net_Node(
            "T_CHILD",
            upstream=prefs.get('NAME'),
            port=prefs.get('MSGPORT'),
            listens = {},
            instance=True
        )

        while True:
            for name, cam in self.cams.items():
                try:
                    frame, timestamp = cam.q.get_nowait()
                    self.node.send(key='CONTINUOUS',
                                   value={cam.name:frame,
                                          'timestamp':timestamp},
                                   repeat=False,
                                   flags={'MINPRINT':True})
                except Empty:
                    pass



    def noop(self):
        # just fitting in with the task structure.
        self.stage_block.clear()
        return {}

    # def start(self):
    #     for cam in self.cams.values():
    #         cam.capture()
    #
    # def stop(self):
    #     for cam in self.cams.values():
    #         cam.release()

class Transformer(Child):

    def __init__(self, transform,
                 operation: str ="trigger",
                 node_id = None,
                 return_id = 'T',
                 return_ip = None,
                 return_port = None,
                 return_key = None,
                 router_port = None,
                 stage_block = None,
                 value_subset=None,
                 forward_id=None,
                 forward_ip=None,
                 forward_port=None,
                 forward_key=None,
                 forward_what='both',
                 **kwargs):
        """

        Args:
            transform:
            operation (str): either

                * "trigger", where the last transform is a :class:`~autopilot.transform.transforms.Condition`
                and a trigger is returned to sender only when the return value of the transformation changes, or
                * "stream", where each result of the transformation is returned to sender

            return_id:
            return_ip:
            return_port:
            return_key:
            router_port (None, int): If not ``None`` (default), spawn the node with a route port to receieve
            stage_block:
            value_subset (str): Optional - subset a value from from a dict/list sent to :meth:`.l_process`
            forward_what (str): one of 'input', 'output', or 'both' (default) that determines what is forwarded
            **kwargs:
        """
        super(Transformer, self).__init__(**kwargs)
        assert operation in ('trigger', 'stream', 'debug')
        self.operation = operation
        self._last_result = None

        if return_key is None:
            self.return_key = self.operation.upper()
        else:
            self.return_key = return_key

        self.return_id = return_id
        self.return_ip = return_ip
        self.return_port = return_port
        if self.return_port is None:
            self.return_port = prefs.get('MSGPORT')
        if node_id is None:
            self.node_id = f"{prefs.get('NAME')}_TRANSFORMER"
        else:
            self.node_id = node_id
        self.router_port = router_port

        self.forward_id = forward_id
        self.forward_ip = forward_ip
        self.forward_port = forward_port
        self.forward_key = forward_key
        self.forward_node = None
        self.forward_what = forward_what

        self.stage_block = stage_block
        self.stages = cycle([self.noop])
        # self.input_q = LifoQueue()
        self.input_q = deque(maxlen=1)
        self.value_subset = value_subset

        self.logger = init_logger(self)

        self.process_thread = threading.Thread(target=self._process, args=(transform,))
        self.process_thread.daemon = True
        self.process_thread.start()

    def noop(self):
        # just fitting in with the task structure.
        self.stage_block.clear()
        return {}



    def _process(self, transform):

        self.transform = autopilot.transform.make_transform(transform)


        self.node = Net_Node(
            self.node_id,
            upstream=self.return_id,
            upstream_ip=self.return_ip,
            port=self.return_port,
            router_port=self.router_port,
            listens = {
                'CONTINUOUS': self.l_process
            },
            instance=False
        )

        if all([x is not None for x in
                (self.forward_id,
                 self.forward_ip,
                 self.forward_key,
                 self.forward_port)]):
            self.forward_node = Net_Node(
                id=self.node_id,
                upstream=self.forward_id,
                upstream_ip=self.forward_ip,
                port=self.forward_port,
                listens={}
            )


        self.node.send(self.return_id, 'STATE', value='READY')

        while True:
            try:
                # value = self.input_q.get_nowait()
                value = self.input_q.popleft()
            # except Empty:
            except IndexError:
                sleep(0.001)
                continue
            result = self.transform.process(value)

            self.node.logger.debug(f'Processed frame, result: {result}')

            if self.operation == "trigger":
                if result != self._last_result:
                    self.node.send(self.return_id, self.return_key, result)
                    if self.forward_node is not None:
                        self.forward(value, result)
                    self._last_result = result

            elif self.operation == 'stream':
                # FIXME: Another key that's not TRIGGER
                self.node.send(self.return_id, self.return_key, result)
                if self.forward_node is not None:
                    self.forward(value, result)

            elif self.operation == 'debug':
                pass


    def l_process(self, value):
        # get array out of value

        # FIXME hack for dlc
        self.node.logger.debug('Received and queued processing!')
        # self.input_q.put_nowait(value['MAIN'])
        if self.value_subset:
            value = value[self.value_subset]
        self.input_q.append(value)

    def forward(self, input=None, output=None):
        if self.forward_what == 'both':
            self.forward_node.send(self.forward_id, self.forward_key, {'input':input,'output':output},flags={'MINPRINT':True,'NOREPEAT':True})
        elif self.forward_what == 'input':
            self.forward_node.send(self.forward_id, self.forward_key, input,flags={'MINPRINT':True,'NOREPEAT':True})
        elif self.forward_what == 'output':
            self.forward_node.send(self.forward_id, self.forward_key, output,flags={'MINPRINT':True,'NOREPEAT':True})
        else:
            raise ValueError("forward_what must be one of 'input', 'output', or 'both'")















import pytest
from pathlib import Path
from autopilot.prefs import _DEFAULTS, Scopes

@pytest.fixture
def default_dirs():
    for k, v in _DEFAULTS.items():
        if v['scope'] == Scopes.DIRECTORY and 'default' in v.keys():
            Path(v['default']).mkdir(parents=True, exist_ok=True)




"""
Classes for network communication.

There are two general types of network objects -

* :class:`autopilot.networking.Station` and its children are independent processes that should only be instantiated once
    per piece of hardware. They are used to distribute messages between :class:`.Net_Node` s,
    forward messages up the networking tree, and responding to messages that don't need any input from
    the :class:`~.pilot.Pilot` or :class:`~.terminal.Terminal`.
* :class:`.Net_Node` is a pop-in networking class that can be given to any other object that
    wants to send or receive messages.

The :class:`~autopilot.networking.Message` object is used to serialize and pass
messages. When sent, messages are ``JSON`` serialized (with some special magic
to compress/encode numpy arrays) and sent as ``zmq`` multipart messages.

Each serialized message, when sent, can have ``n`` frames of the format::

    [hop_0, hop_1, ... hop_n, final_recipient, serialized_message]

Or, messages can have multiple "hops" (a typical message will have one 'hop' specified
by the ``to`` field), the second to last frame is always the final intended recipient,
and the final frame is the serialized message. Note that the ``to`` field of a
:class:`~autopilot.networking.Message` object will always be the final recipient
even if a list is passed for ``to`` when sending. This lets :class:`~.networking.Station`
objects efficiently forward messages without deserializing them at every hop.
"""


import base64

import blosc

from autopilot.networking.station import Station, Terminal_Station, Pilot_Station
from autopilot.networking.node import Net_Node
from autopilot.networking.message import Message


def serialize_array(array):
    """
    Pack an array with :func:`blosc.pack_array` and serialize with :func:`base64.b64encode`

    Args:
        array (:class:`numpy.ndarray`): Array to serialize

    Returns:
        dict: {'NUMPY_ARRAY': base-64 encoded, blosc-compressed array.}
    """
    compressed = base64.b64encode(blosc.pack_array(array)).decode('ascii')
    return {'NUMPY_ARRAY': compressed}


"""
Utility functions for dealing with the wiki (https://wiki.auto-pi-lot.com).

See the docstrings of the :func:`.ask` function, as well as the :ref:`guide_wiki_plugins` section
in the user guide for use.
"""
from typing import Union, Optional, List
import requests
import pdb
import json
from autopilot.utils.common import find_key_value

WIKI_URL = "https://wiki.auto-pi-lot.com/"

def ask(
        filters:Union[List[str],str],
        properties:Union[None,List[str],str]=None) -> List[dict]:
    """
    Perform an API call to the wiki using the `ask API <https://www.semantic-mediawiki.org/wiki/Help:API:ask>`_
    and simplify to a list of dictionaries

    Args:
        filters (list, str): A list of strings or a single string of semantic
            mediawiki formatted property filters. See :func:`.make_ask_string` for more information
        properties (None, list, str): Properties to return from filtered pages,
            See :func:`.make_ask_string` for more information

    Returns:

    """
    query_str = make_ask_string(filters=filters, properties=properties, full_url=True)
    result = requests.get(query_str)
    entries = result.json()['query']['results']

    # unnest 'printouts' and convert from a nested list of dicts with the top
    # dict as the name (eg {'entry_name':{prop1:'prop1'}} to {'name': 'entry_name', 'prop1':...}
    unnested = []
    for entry in entries:
        entry_name = list(entry.keys())[0]
        nested_entry = entry[entry_name]
        unnest_entry = _clean_smw_result(nested_entry)
        unnested.append(unnest_entry)

    return unnested

def browse(
        search:str,
        browse_type:str="page",
        params:Optional[dict]=None
):
    """
    Use the `browse <https://www.semantic-mediawiki.org/wiki/Help:API:smwbrowse>`_ api of the wiki  to
    search for specific pages, properties, and so on.

    Args:
        search (str): the search string! ``*`` can be used as a wildcard.
        browse_type (str): The kind of browsing we're doing, one of:

            * page
            * subject
            * property
            * pvalue
            * category
            * concept

        params (dict): Additional params for the browse given as a dictionary, see `the smw docs <https://www.semantic-mediawiki.org/wiki/Help:API:smwbrowse>`_
            for usage.

    Returns:
        dict, list of dicts of results
    """

    browse_str = make_browse_string(search=search, browse_type=browse_type, params=params, full_url=True)
    result = requests.get(browse_str)


def _clean_smw_result(nested_entry:dict) -> dict:
    # unnest entries that are [[Has type::page]] and thus have extra metadata
    unnest_entry = {}
    printouts = nested_entry.get('printouts', {})
    if len(printouts)>0:
        for k, v in printouts.items():
            if isinstance(v, list) and len(v) > 1:
                unnest_entry[k] = []
                for subv in v:
                    if isinstance(subv, dict) and 'fulltext' in subv.keys():
                        subv = subv['fulltext']
                    unnest_entry[k].append(subv)
                unnest_entry[k] = sorted(unnest_entry[k])
            elif isinstance(v, list) and len(v) == 1:
                unnest_entry[k] = v[0]
                if isinstance(unnest_entry[k], dict) and 'fulltext' in unnest_entry[k].keys():
                    unnest_entry[k] = unnest_entry[k]['fulltext']
            else:
                unnest_entry[k] = v

    unnest_entry['name'] = nested_entry['fulltext']
    unnest_entry['url'] = nested_entry['fullurl']
    return unnest_entry


def make_ask_string(filters:Union[List[str], str],
                    properties:Union[None,List[str],str]=None,
                    full_url:bool = True) -> str:
    """
    Create a query string to request semantic information from the Autopilot wiki

    Args:
        filters (list, str): A list of strings or a single string of semantic
            mediawiki formatted property filters, eg ``"[[Category:Hardware]]"``
            or ``"[[Has Contributor::sneakers-the-rat]]"``. Refer to the
            `semantic mediawiki documentation <https://www.semantic-mediawiki.org/wiki/Help:Selecting_pages>`_
            for more information on syntax
        properties (None, list, str): Properties to return from filtered pages,
            see the `available properties <https://wiki.auto-pi-lot.com/index.php/Special:Properties>`_
            on the wiki and the `semantic mediawiki documentation <https://www.semantic-mediawiki.org/wiki/Help:Selecting_pages>`_
            for more information on syntax. If ``None`` (default), just return
            the names of the pages
        full_url (bool): If ``True`` (default), prepend ``f'{WIKI_URL}api.php?action=ask&query='``
            to the returned string to make it `ready for an API call <https://www.semantic-mediawiki.org/wiki/Help:API:ask>`_

    Returns:
        str: the formatted query string
    """
    # combine the components, separated by pipes or pip question marks as the case may be
    if isinstance(filters, str):
        filters = [filters]

    if len(filters)==0:
        raise ValueError(f'You need to provide at least one filter! Cant get the whole wiki!')

    query_str = "|".join(filters)

    if isinstance(properties, str):
        properties = [properties]
    elif properties is None:
        properties = []

    if len(properties)>0:
        # double join with ?| so it goes between
        # all the properties *and* between filters and
        query_str = "|?".join((
            query_str,
            "|?".join(properties)
        ))

    # add api call boilerplate and URI-encode
    query_str = requests.utils.quote(query_str) + "&format=json&api_version=3"

    if full_url:
        return f"{WIKI_URL}api.php?action=ask&query=" + query_str
    else:
        return query_str

def make_browse_string(search, browse_type='page', params=None, full_url:bool=True):

    if params is None:
        params = {}
    params['search'] = search

    browse_str = f"&browse={browse_type}&params={json.dumps(params)}&format=json"
    if full_url:
        return f"{WIKI_URL}api.php?action=smwbrowse" + browse_str
    else:
        return browse_str







import numpy as np
import pdb

from autopilot.transform.geometry import Spheroid, _ellipsoid_func

n_samples = 100

def test_spheroid_init():
    # test that the dang thing initializes as expected with no params
    sphere_1 = Spheroid()
    assert sphere_1._scale is None
    assert sphere_1._offset_source is None
    assert sphere_1._offset_target is None

    # test that when given explicit source that arrays are generated correctly
    sphere_2 = Spheroid(target=(1,2,3,0,1,2),
                        source=(2,4,6,1,3,5))
    assert np.array_equal(sphere_2._scale, np.array((0.5,0.5,0.5)))
    assert np.array_equal(sphere_2._offset_source, np.array((1, 3, 5)))
    assert np.array_equal(sphere_2._offset_target, np.array((0,1,2)))

def test_spheroid_generate_fit():
    for i in range(n_samples):
        # parameterize a target spheroid
        target = ((np.random.rand()+5)*5,
                  (np.random.rand()+5)*5,
                  (np.random.rand()+5)*5,
                  (np.random.rand()-0.5)*20,
                  (np.random.rand()-0.5)*20,
                  (np.random.rand()-0.5)*20)
        # make a spheroid to generate points
        sphere_generator = Spheroid(target=target)

        # test that the generation is good
        noise = np.random.rand()
        max_distance = np.sqrt((noise**2)*3)
        pts = sphere_generator.generate(1000, which="target",noise=noise)
        ellipsoid_fn_out = _ellipsoid_func(pts, *target)
        assert np.all(ellipsoid_fn_out-1 < max_distance)

        # test that fit recovers the target
        sphere_generator.fit(pts, bounds=((0,0,0,-50,-50,-50),(50,50,50,50,50,50)))
        for t_param, fit_param in zip(target, sphere_generator.source):
            assert np.abs(t_param - fit_param) < max_distance

def test_spheroid_process():
    for i in range(n_samples):
        # test that for a known target and source, the target is correctly
        # recovered from points on the source spheroid

        target = ((np.random.rand() + 1) * 10,
                  (np.random.rand() + 1) * 10,
                  (np.random.rand() + 1) * 10,
                  (np.random.rand() - 0.5) * 20,
                  (np.random.rand() - 0.5) * 20,
                  (np.random.rand() - 0.5) * 20)

        source = ((np.random.rand() + 1) * 10,
                  (np.random.rand() + 1) * 10,
                  (np.random.rand() + 1) * 10,
                  (np.random.rand() - 0.5) * 20,
                  (np.random.rand() - 0.5) * 20,
                  (np.random.rand() - 0.5) * 20)

        sphere = Spheroid(target=target, source=source)
        pts = sphere.generate(1000, which="source", noise=0)
        pts_tfm = sphere.process(pts)

        pts_test = _ellipsoid_func(pts, *source)
        pts_tfm_test = _ellipsoid_func(pts_tfm, *target)


        assert np.allclose(pts_test, np.ones(1000))
        assert np.allclose(pts_tfm_test, np.ones(1000))



"""
Tools to visulize data after collection.

Warning:
    this module is unfinished, so it is undocumented.
"""

# renders a standalone webpage with bokeh of trial data for all subjects in the data folder
import sys
import os

import autopilot.utils
import autopilot.utils.common

sys.path.append(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))))

from glob import glob
import argparse
from bokeh.plotting import figure
from bokeh.io import show
from bokeh.models import ColumnDataSource, Legend, LegendItem, Span
from bokeh.layouts import gridplot
from bokeh.transform import factor_cmap
from bokeh.palettes import Spectral10
from tqdm import tqdm
from autopilot.core import subject
import colorcet as cc
import numpy as np
import json


def load_subject_data(data_dir, subject_name, steps=True, grad=True):

    # pilot_db_fn = [fn for fn in os.listdir(data_dir) if fn == 'pilot_db.json'][0]
    # pilot_db_fn = os.path.join(data_dir, pilot_db_fn)
    pilot_db = autopilot.utils.common.load_pilotdb(reverse=True)

    # find pilot for subject
    pilot_name = pilot_db[subject_name]

    amus = subject.Subject(subject_name, dir=data_dir)

    step_data = None
    grad_data = None

    if steps:
        step_data = amus.get_trial_data()
        step_data['subject'] = subject_name
        step_data['pilot'] = pilot_name

    if grad:
        # get historical graduation data
        try:
            grad_data = amus.get_step_history()
        except:
            grad_data = amus.get_step_history(use_history=False)

        grad_data['subject'] = subject_name
        grad_data['pilot'] = pilot_name

    return step_data, grad_data







def load_subject_dir(data_dir, steps=True, grad=True, which = None):
    """
    Args:
        data_dir (str): A path to a directory with :class:`~.core.subject.Subject` style hdf5 files
        steps (bool): Whether to return full trial-level data for each step
        grad (bool): Whether to return summarized step graduation data.
        which (list): A list of subjects to subset the loaded subjects to

    """

    subject_fn = [os.path.splitext(fn)[0] for fn in os.listdir(data_dir) if fn.endswith('.h5')]

    if isinstance(which, list):
        subject_fn = [fn for fn in subject_fn if (fn in which) or (fn.rstrip('.h5') in which)]

    all_mice_steps = None
    all_mice_grad = None

    for subject_name in tqdm(subject_fn):
        subject_name = os.path.splitext(subject_name)[0]
        step_data, grad_data = load_subject_data(data_dir, subject_name, steps, grad)

        if step_data is not None:
            if all_mice_steps is not None:
                all_mice_steps = all_mice_steps.append(step_data)
            else:
                all_mice_steps = step_data

        if grad_data is not None:
            if all_mice_grad is not None:
                all_mice_grad = all_mice_grad.append(grad_data)
            else:
                all_mice_grad = grad_data

    return all_mice_steps, all_mice_grad




def step_viewer(grad_data):
    mice = sorted(grad_data['subject'].unique())
    palette = [cc.rainbow[i] for i in range(len(grad_data['pilot'].unique()))]

    current_step = grad_data.groupby('subject').last().reset_index()
    current_step = current_step[['subject', 'step_n', 'pilot']]

    pilots = current_step['pilot'].unique()
    pilot_colors = {p:palette[i] for i,p in enumerate(pilots) }
    pilot_colors = [pilot_colors[p] for p in current_step['pilot']]
    current_step['colors'] = pilot_colors




    p = figure(x_range=current_step['subject'].unique(),title='Subject Steps',
               plot_height=600,
               plot_width=1000)
    p.xaxis.major_label_orientation = np.pi / 2
    bars = p.vbar(x='subject', top='step_n', width=0.9,
           fill_color=factor_cmap('pilot', palette=Spectral10, factors=pilots),
           legend='pilot',
           source=ColumnDataSource(current_step))
    p.legend.location = 'top_center'
    p.legend.orientation = 'horizontal'
    #p.add_layout(legend,'below')

    show(p)




def trial_viewer(step_data, roll_type = "ewm", roll_span=100, bar=False):
    """
    Args:
        bar:
        roll_span:
        roll_type:
        step_data:
    """
    step_data.loc[step_data['response'] == 'L','response'] = 0
    step_data.loc[step_data['response'] == 'R','response'] = 1
    step_data.loc[step_data['target'] == 'L','target'] = 0
    step_data.loc[step_data['target'] == 'R','target'] = 1

    palette = [cc.rainbow[i] for i in range(len(step_data['subject'].unique()))]
    palette = [cc.rainbow[i*15] for i in range(5)]

    mice = sorted(step_data['subject'].unique())
    current_step = step_data.groupby('subject').last().reset_index()
    current_step = current_step[['subject','step']]

    plots = []
    p = figure(x_range=step_data['subject'].unique(),title='Subject Steps',
               plot_height=200)
    p.xaxis.major_label_orientation = np.pi / 2
    p.vbar(x=current_step['subject'], top=current_step['step'], width=0.9)
    plots.append(p)
    for i, (mus, group) in enumerate(step_data.groupby('subject')):
        if roll_type == "ewm":
            meancx = group['correct'].ewm(span=roll_span,ignore_na=True).mean()
        else:
            meancx = group['correct'].rolling(window=roll_span).mean()

        title_str = "{}, step: {}".format(mus, group.step.iloc[-1])

        p = figure(plot_height=100,y_range=(0,1),title=title_str)

        if bar:
            hline = Span(location=bar, dimension="width", line_color='red', line_width=1)
            p.renderers.append(hline)

        p.line(group['trial_num'], meancx, color=palette[group['step'].iloc[0]-1])
        plots.append(p)
    grid = gridplot(plots, ncols=1)
    show(grid)

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description="Visualize Trial Data")
    parser.add_argument('-d', '--dir', help="Data directory")
    parser.add_argument('-t', '--type', help="Type of plot? s=steps, g=graduation")
    parser.add_argument('-w', '--window', help="Window of trials to roll over in step plot")
    parser.add_argument('-r', '--roll', help="Type of roll, ewm=exponentially weighted mean, anything else = equally weighted")
    parser.add_argument('-b', '--bar', help="position to draw horizontal bar")
    args = parser.parse_args()

    if not args.dir:
        data_dir = '/usr/autopilot/data'

        if not os.path.exists(data_dir):
            raise Exception("No directory file passed, and default location doesn't exist")

        raise Warning('No directory passed, loading from default location. Should pass explicitly with -d')
    else:
        data_dir = args.dir

    # TODO Make arg
    active_mice = autopilot.utils.common.list_subjects()








    if not args.type:
        do_type = 'g' # raduation, aka what stage they're on
    else:
        do_type = str(args.type)

    if args.bar:
        bar_pos = float(args.bar)
    else:
        bar_pos = False

    if do_type == 'g':
        # load subject data
        print('Doing graduation plot,\nloading subject data...')
        # step_data, grad_data = load_subject_data(data_dir)
        _, grad_data = load_subject_dir(data_dir, steps=False, grad=True, which=active_mice)

        step_viewer(grad_data)
    elif do_type == "s":
        # load subject data
        print('Doing step plot,\nloading subject data...')
        # step_data, grad_data = load_subject_data(data_dir)
        step_data, _ = load_subject_dir(data_dir, steps=True, grad=False, which=active_mice)

        if args.window:
            window = int(args.window)
        else:
            window = 100

        if args.roll:
            roll_type = str(args.roll)
        else:
            roll_type = "ewm"



        trial_viewer(step_data, roll_span=window, roll_type=roll_type, bar=bar_pos)












import pytest
import pdb
import inspect
from typing import Tuple
from pathlib import Path
import autopilot
from autopilot import prefs
from autopilot.hardware import Hardware
from autopilot.utils.registry import get
from autopilot.utils.plugins import _IMPORTED
from tests.fixtures import default_dirs


@pytest.fixture
def hardware_plugin(default_dirs) -> Tuple[Path, str]:
    """
    Make a basic plugin that inherits from the Hardware class,
    clean it up on exit

    Returns:
        Path: path to created plugin file
    """
    plugin_dir = Path(prefs.get('PLUGINDIR'))
    prefs.set('AUTOPLUGIN', True)
    hw_fn = plugin_dir / "test_hardware_plugin.py"
    class_name = "Test_Hardware_Plugin"
    hardware_plugin_test = f"""
from autopilot.hardware import Hardware

class {class_name}(Hardware):
    def __init__(self, *args, **kwargs):
        name="Test_Hardware_Plugin"
        super(Test_Hardware_Plugin, self).__init__(name=name, *args, **kwargs)
    
    def im_alive(self):
        return True
        
    def release(self):
        pass
        
    """

    with open(hw_fn, 'w') as hw_file:
        hw_file.write(hardware_plugin_test)
    _IMPORTED.value = False

    yield (hw_fn, class_name)

    # cleanup
    if hw_fn.exists():
        hw_fn.unlink()

def test_hardware_plugin(hardware_plugin):
    """
    A subclass of :class:`autopilot.hardware.Hardware` in the ``PLUGINDIR`` can be
    accessed with :func:`autopilot.get`.

    For example, for the following class declared in some ``.py`` file in the plugin dir::

        from autopilot.hardware import Hardware

        class Test_Hardware_Plugin(Hardware):
            def __init__(self, *args, **kwargs):
                super(Test_Hardware_Plugin, self).__init__(*args, **kwargs)

            def release(self):
                pass

    one would be able to access it throughout autopilot with::

        autopilot.get('hardware', 'Test_Hardware_Plugin')
        # or
        autopilot.get_hardware('Test_Hardware_Plugin')

    """
    hw_path, class_name = hardware_plugin

    assert hw_path.exists()

    plugin_class = autopilot.get('hardware', class_name)

    assert inspect.isclass(plugin_class)
    assert issubclass(plugin_class, Hardware)
    assert plugin_class().im_alive()

def test_autoplugin():
    """
    the :func:`autopilot.utils.registry.get` function should automatically load
    plugins if the pref ``AUTOPLUGIN`` is ``True`` and the ``plugins`` argument is True
    """
    pass







"""

"""

import os
import sys
import datetime
import logging
import argparse
import threading
import time
import socket
import json
import base64
import subprocess
import warnings
import numpy as np
import pandas as pd
from scipy.stats import linregress

import tables
warnings.simplefilter('ignore', category=tables.NaturalNameWarning)

import autopilot
from autopilot import prefs
from autopilot.core.loggers import init_logger

if __name__ == '__main__':
    # Parse arguments - this should have been called with a .json prefs file passed
    # We'll try to look in the default location first
    parser = argparse.ArgumentParser(description="Run an autopilot")
    parser.add_argument('-f', '--prefs', help="Location of .json prefs file (created during setup_autopilot.py)")
    args = parser.parse_args()

    if not args.prefs:
        prefs_file = '/usr/autopilot/prefs.json'

        if not os.path.exists(prefs_file):
            raise Exception("No Prefs file passed, and file not in default location")

        raise Warning('No prefs file passed, loaded from default location. Should pass explicitly with -p')

    else:
        prefs_file = args.prefs

    prefs.init(prefs_file)

    if prefs.get('AUDIOSERVER') or 'AUDIO' in prefs.get('CONFIG'):
        if prefs.get('AUDIOSERVER') == 'pyo':
            from autopilot.stim.sound import pyoserver
        else:
            from autopilot.stim.sound import jackclient

from autopilot.networking import Message, Net_Node, Pilot_Station
from autopilot import external
from autopilot.hardware import gpio


########################################

class Pilot:
    """
    Drives the Raspberry Pi

    Coordinates the hardware and networking objects to run tasks.

    Typically used with a connection to a :class:`.Terminal` object to
    coordinate multiple subjects and tasks, but a high priority for future releases
    is to do the (trivial amount of) work to make this class optionally
    standalone.

    Called as a module with the -f flag to give the location of a prefs file, eg::

        python pilot.py -f prefs_file.json

    if the -f flag is not passed, looks in the default location for prefs
    (ie. `/usr/autopilot/prefs.json`)

    Needs the following prefs (typically established by :mod:`.setup.setup_pilot`):

    * **NAME** - The name used by networking objects to address this Pilot
    * **BASEDIR** - The base directory for autopilot files (/usr/autopilot)
    * **PUSHPORT** - Router port used by the Terminal we connect to.
    * **TERMINALIP** - IP Address of our upstream Terminal.
    * **MSGPORT** - Port used by our own networking object
    * **HARDWARE** - Any hardware and its mapping to GPIO pins. No pins are required to be set, instead each
      task defines which pins it needs. Currently the default configuration asks for

        * POKES - :class:`.hardware.Beambreak`
        * LEDS - :class:`.hardware.LED_RGB`
        * PORTS - :class:`.hardware.Solenoid`

    * **AUDIOSERVER** - Which type, if any, audio server to use (`'jack'`, `'pyo'`, or `'none'`)
    * **NCHANNELS** - Number of audio channels
    * **FS** - Sampling rate of audio output
    * **JACKDSTRING** - string used to start the jackd server, see `the jack manpages <https://linux.die.net/man/1/jackd>`_ eg::

        jackd -P75 -p16 -t2000 -dalsa -dhw:sndrpihifiberry -P -rfs -n3 -s &

    * **PIGPIOMASK** - Binary mask of pins for pigpio to control, see `the pigpio docs <http://abyz.me.uk/rpi/pigpio/pigpiod.html>`_ , eg::

        1111110000111111111111110000

    * **PULLUPS** - Pin (board) numbers to pull up on boot
    * **PULLDOWNS** - Pin (board) numbers to pull down on boot.

    Attributes:
        name (str): The name used to identify ourselves in :mod:`.networking`
        task (:class:`.tasks.Task`): The currently instantiated task
        running (:class:`threading.Event`): Flag used to control task running state
        stage_block (:class:`threading.Event`): Flag given to a task to signal when task stages finish
        file_block (:class:`threading.Event`): Flag used to wait for file transfers
        state (str): 'RUNNING', 'STOPPING', 'IDLE' - signals what this pilot is up to
        pulls (list): list of :class:`~.hardware.Pull` objects to keep pins pulled up or down
        server: Either a :func:`~.sound.pyoserver.pyo_server` or :class:`~.jackclient.JackClient` , sound server.
        node (:class:`.networking.Net_Node`): Our Net_Node we use to communicate with our main networking object
        networking (:class:`.networking.Pilot_Station`): Our networking object to communicate with the outside world
        ip (str): Our IPv4 address
        listens (dict): Dictionary mapping message keys to methods used to process them.
        logger (:class:`logging.Logger`): Used to log messages and network events.
    """

    logger = None


    # Events for thread handling
    running = None
    stage_block = None
    file_block = None
    quitting = None
    """mp.Event to signal when process is quitting"""

    # networking - our internal and external messengers
    node = None
    networking = None

    # audio server
    server = None

    def __init__(self, splash=True):

        if splash:
            with open(os.path.join(os.path.dirname(os.path.dirname(__file__)), 'setup', 'welcome_msg.txt'), 'r') as welcome_f:
                welcome = welcome_f.read()
                print('')
                for line in welcome.split('\n'):
                    print(line)
                print('')
                sys.stdout.flush()

        self.name = prefs.get('NAME')
        if prefs.get('LINEAGE') == "CHILD":
            self.child = True
            self.parentid = prefs.get('PARENTID')
        else:
            self.child = False
            self.parentid = 'T'

        self.logger = init_logger(self)
        self.logger.debug('pilot logger initialized')

        # Locks, etc. for threading
        self.running = threading.Event() # Are we running a task?
        self.stage_block = threading.Event() # Are we waiting on stage triggers?
        self.file_block = threading.Event() # Are we waiting on file transfer?
        self.quitting = threading.Event()
        self.quitting.clear()

        # init pigpiod process
        self.init_pigpio()

        # Init audio server
        if prefs.get('AUDIOSERVER') or 'AUDIO' in prefs.get('CONFIG'):
            self.init_audio()


        # Init Station
        # Listen dictionary - what do we do when we receive different messages?
        self.listens = {
            'START': self.l_start, # We are being passed a task and asked to start it
            'STOP' : self.l_stop, # We are being asked to stop running our task
            'PARAM': self.l_param, # A parameter is being changed
            'CALIBRATE_PORT': self.l_cal_port, # Calibrate a water port
            'CALIBRATE_RESULT': self.l_cal_result, # Compute curve and store result
            'BANDWIDTH': self.l_bandwidth, # test our bandwidth
            'STREAM_VIDEO': self.l_stream_video
        }

        # spawn_network gives us the independent message-handling process
        self.networking = Pilot_Station()
        self.networking.start()
        self.node = Net_Node(id = "_{}".format(self.name),
                             upstream = self.name,
                             port = prefs.get('MSGPORT'),
                             listens = self.listens,
                             instance=False)
        self.logger.debug('pilot networking initialized')

        # if we need to set pins pulled up or down, do that now
        self.pulls = []
        if prefs.get( 'PULLUPS'):
            for pin in prefs.get('PULLUPS'):
                self.pulls.append(gpio.Digital_Out(int(pin), pull='U', polarity=0))
        if prefs.get( 'PULLDOWNS'):
            for pin in prefs.get('PULLDOWNS'):
                self.pulls.append(gpio.Digital_Out(int(pin), pull='D', polarity=1))

        # store some hardware we use outside of a task
        self.hardware = {}


        self.logger.debug('pullups and pulldowns set')
        # check if the calibration file needs to be updated


        # Set and update state
        self.state = 'IDLE' # or 'Running'
        self.update_state()

        # Since we're starting up, handshake to introduce ourselves
        self.ip = self.get_ip()
        self.handshake()
        self.logger.debug('handshake sent')



        #self.blank_LEDs()

        # TODO Synchronize system clock w/ time from terminal.



    #################################################################
    # Station
    #################################################################

    def get_ip(self):
        """
        Get our IP
        """

        # shamelessly stolen from https://www.w3resource.com/python-exercises/python-basic-exercise-55.php
        # variables are badly named because this is just a rough unwrapping of what was a monstrous one-liner

        # get ips that aren't the loopback
        unwrap00 = [ip for ip in socket.gethostbyname_ex(socket.gethostname())[2] if not ip.startswith("127.")][:1]
        # ???
        unwrap01 = [[(s.connect(('8.8.8.8', 53)), s.getsockname()[0], s.close()) for s in [socket.socket(socket.AF_INET, socket.SOCK_DGRAM)]][0][1]]

        unwrap2 = [l for l in (unwrap00,unwrap01) if l][0][0]

        return unwrap2

    def handshake(self):
        """
        Send the terminal our name and IP to signal that we are alive
        """
        # send the terminal some information about ourselves

        # TODO: Report any calibrations that we have

        hello = {'pilot':self.name, 'ip':self.ip, 'state':self.state, 'prefs': prefs.get()}

        self.node.send(self.parentid, 'HANDSHAKE', value=hello)

    def update_state(self):
        """
        Send our current state to the Terminal,
        our Station object will cache this and will handle any
        future requests.
        """
        self.node.send(self.name, 'STATE', self.state, flags={'NOLOG':True})

    def l_start(self, value):
        """
        Start running a task.

        Get the task object by using `value['task_type']` to select from
        :func:`autopilot.get_task()` , then feed the rest of `value` as kwargs
        into the task object.

        Calls :meth:`.autopilot.run_task` in a new thread

        Args:
            value (dict): A dictionary of task parameters
        """
        # TODO: If any of the sounds are 'file,' make sure we have them. If not, request them.
        # Value should be a dict of protocol params
        # The networking object should have already checked that we have all the files we need

        if self.state == "RUNNING" or self.running.is_set():
            self.logger.warning("Asked to a run a task when already running")
            return

        self.state = 'RUNNING'
        self.running.set()
        try:
            # Get the task object by its type
            if 'child' in value.keys():
                task_class = autopilot.get('children', value['task_type'])
            else:
                task_class = autopilot.get_task(value['task_type'])
            # Instantiate the task
            self.stage_block.clear()

            # Make a group for this subject if we don't already have one
            self.subject = value['subject']
            prefs.set('SUBJECT', self.subject)



            # Run the task and tell the terminal we have
            # self.running.set()
            threading.Thread(target=self.run_task, args=(task_class, value)).start()


            self.update_state()
        except Exception as e:
            self.state = "IDLE"
            self.logger.exception("couldn't start task: {}".format(e))

        # TODO: Send a message back to the terminal with the runtime if there is one so it can handle timed stops

    def l_stop(self, value):
        """
        Stop the task.

        Clear the running event, set the stage block.

        TODO:
            Do a coherence check between our local file and the Terminal's data.

        Args:
            value: ignored
        """
        # Let the terminal know we're stopping
        # (not stopped yet because we'll still have to sync data, etc.)
        self.state = 'STOPPING'
        self.update_state()

        # We just clear the stage block and reset the running flag here
        # and call the cleanup routine from run_task so it can exit cleanly
        self.running.clear()
        self.stage_block.set()



        # TODO: Cohere here before closing file
        if hasattr(self, 'h5f'):
            self.h5f.close()

        self.state = 'IDLE'
        self.update_state()

    def l_param(self, value):
        """
        Change a task parameter mid-run

        Warning:
            Not Implemented

        Args:
            value:
        """
        pass

    def l_cal_port(self, value):
        """
        Initiate the :meth:`.calibrate_port` routine.

        Args:
            value (dict): Dictionary of values defining the port calibration to be run, including
                - ``port`` - which port to calibrate
                - ``n_clicks`` - how many openings should be performed
                - ``open_dur`` - how long the valve should be open
                - ``iti`` - 'inter-trial interval`, or how long should we wait between valve openings.

        """
        port = value['port']
        n_clicks = value['n_clicks']
        open_dur = value['dur']
        iti = value['click_iti']

        threading.Thread(target=self.calibrate_port,args=(port, n_clicks, open_dur, iti)).start()

    def calibrate_port(self, port_name, n_clicks, open_dur, iti):
        """
        Run port calibration routine

        Open a :class:`.hardware.gpio.Solenoid` repeatedly,
        measure volume of water dispersed, compute lookup table mapping
        valve open times to volume.

        Continuously sends progress of test with ``CAL_PROGRESS`` messages

        Args:
            port_name (str): Port name as specified in ``prefs``
            n_clicks (int): number of times the valve should be opened
            open_dur (int, float): how long the valve should be opened for in ms
            iti (int, float): how long we should :func:`~time.sleep` between openings

        """
        pin_num = prefs.get('HARDWARE')['PORTS'][port_name]
        port = gpio.Solenoid(pin_num, duration=int(open_dur))
        msg = {'click_num': 0,
               'pilot': self.name,
               'port': port_name
               }

        iti = float(iti)/1000.0

        cal_name = "Cal_{}".format(self.name)

        for i in range(int(n_clicks)):
            port.open()
            msg['click_num'] = i + 1
            self.node.send(to=cal_name, key='CAL_PROGRESS',
                           value= msg)
            time.sleep(iti)

        port.release()

    def l_cal_result(self, value):
        """
        Save the results of a port calibration

        """

        # files for storing raw and fit calibration results
        cal_fn = os.path.join(prefs.get('BASEDIR'), 'port_calibration.json')

        if os.path.exists(cal_fn):
            try:
                with open(cal_fn, 'r') as cal_file:
                    calibration = json.load(cal_file)
            except ValueError:
                # usually no json can be decoded, that's fine calibrations aren't expensive
                calibration = {}
        else:
            calibration = {}

        for port, results in value.items():
            if port in calibration.keys():
                calibration[port].extend(results)
            else:
                calibration[port] = results

        with open(cal_fn, 'w+') as cal_file:
            json.dump(calibration, cal_file)

    def l_bandwidth(self, value):
        """
        Send messages with a poissonian process according to the settings in value
        """
        #turn off logging for now
        self.networking.logger.setLevel(logging.ERROR)
        self.node.logger.setLevel(logging.ERROR)

        n_msg = int(value['n_msg'])
        rate = float(value['rate'])
        payload = int(value['payload'])
        confirm = bool(value['confirm'])

        payload = np.zeros(payload*1024, dtype=np.bool)
        payload_size = sys.getsizeof(payload)

        message = {
            'pilot': self.name,
            'payload': payload,
        }

        # make a fake message to test how large the serialized message is
        test_msg = Message(to='bandwith', key='BANDWIDTH_MSG', value=message, repeat=confirm, flags={'MINPRINT':True},
                           id="test_message", sender="test_sender")
        msg_size = sys.getsizeof(test_msg.serialize())

        message['message_size'] = msg_size
        message['payload_size'] = payload_size

        if rate > 0:
            spacing = 1.0/rate
        else:
            spacing = 0

        # wait for half a second to let the terminal get messages out
        time.sleep(0.25)

        if spacing > 0:
            last_message = time.perf_counter()
            for i in range(n_msg):
                message['n_msg'] = i
                message['timestamp'] = datetime.datetime.now().isoformat()
                self.node.send(to='bandwidth',key='BANDWIDTH_MSG',
                               value=message, repeat=confirm, flags={'MINPRINT':True})
                this_message = time.perf_counter()
                waitfor = np.clip(spacing-(this_message-last_message), 0, spacing)

                #time.sleep(np.random.exponential(1.0/rate))
                # just do linear spacing lol.

                time.sleep(waitfor)
                last_message = time.perf_counter()
        else:
            for i in range(n_msg):
                message['n_msg'] = i
                message['timestamp'] = datetime.datetime.now().isoformat()
                self.node.send(to='bandwidth',key='BANDWIDTH_MSG',
                               value=message, repeat=confirm, flags={'MINPRINT':True})

        self.node.send(to='bandwidth',key='BANDWIDTH_MSG', value={'pilot':self.name, 'test_end':True,
                                                                  'rate': rate, 'payload':payload,
                                                                  'n_msg':n_msg, 'confirm':confirm},
                       flags={'MINPRINT':True})

        #self.networking.set_logging(True)
        #self.node.do_logging.set()

    def l_stream_video(self, value):
        """
        Start or stop video streaming

        Args:
            value (dict): a dictionary of the form::

                {
                    'starting': bool, # whether we're starting (True) or stopping
                    'camera': str, # the camera to start/stop, of form 'group.camera_id'
                    'stream_to': node id that the camera should send to
                }
        """

        starting = value.get('starting', False)
        camera = value.get('camera', None)
        stream_to = value.get('stream_to', None)

        if camera is None or stream_to is None:
            self.logger.exception('Need a camera and a place to stream it to!')
            return

        try:
            cam_group, cam_id = camera.split('.')
        except ValueError:
            self.logger.exception(f'Expected camera id in form group.camera_id, got {camera}')
            return

        if starting:
            if cam_group in prefs.get('HARDWARE') and cam_id in prefs.get('HARDWARE')[cam_group]:
                cam_prefs = prefs.get('HARDWARE')[cam_group][cam_id]
                cam_obj = get_hardware_class(cam_prefs['type'])(**cam_prefs)
                if cam_group not in self.hardware.keys():
                    self.hardware[cam_group] = {}
                self.hardware[cam_group][cam_id] = cam_obj

                cam_obj.stream(to=stream_to, min_size=1)
                cam_obj.capture()
                self.logger.info(f'Starting to stream video from {camera} to {stream_to}')

            else:
                self.logger.exception(f'No camera in group {cam_group} and id {cam_id} is configured in prefs')

        else:
            # stopping!
            if cam_group in self.hardware.keys() and cam_id in self.hardware[cam_group]:
                cam_obj = self.hardware[cam_group][cam_id]
                cam_obj.stop()
                cam_obj.release()
                del self.hardware[cam_group][cam_id]
                self.logger.info(f'Stopped streaming camera {camera}')
            else:
                self.logger.exception(f'No camera was capturing with group {cam_group} and id {cam_id}, have hardware {self.hardware}')



    def calibration_curve(self, path=None, calibration=None):
        """
        # compute curve to compute duration from desired volume

        Args:
            calibration:
            path: If present, use calibration file specified, otherwise use default.
        """

        lut_fn = os.path.join(prefs.get('BASEDIR'), 'port_calibration_fit.json')

        if not calibration:
            # if we weren't given calibration results, load them
            if path:
                open_fn = path
            else:
                open_fn = os.path.join(prefs.get('BASEDIR'), "port_calibration.json")

            with open(open_fn, 'r') as open_f:
                calibration = json.load(open_f)

        luts = {}
        for port, samples in calibration.items():
            sample_df = pd.DataFrame(samples)
            # TODO: Filter for only most recent timestamps

            # volumes are saved in mL because of how they are measured, durations are stored in ms
            # but reward volumes are typically in the uL range, so we make the conversion
            # by multiplying by 1000
            line_fit = linregress((sample_df['vol']/sample_df['n_clicks'])*1000., sample_df['dur'])
            luts[port] = {'intercept': line_fit.intercept,
                          'slope': line_fit.slope}

        # write to file, overwriting any previous
        with open(lut_fn, 'w') as lutf:
            json.dump(luts, lutf)











    #################################################################
    # Hardware Init
    #################################################################

    def init_pigpio(self):
        try:
            self.pigpiod = external.start_pigpiod()
            self.logger.debug('pigpio daemon started')
        except ImportError as e:
            self.pigpiod = None
            self.logger.exception(e)

    def init_audio(self):
        """
        Initialize an audio server depending on the value of
        `prefs.get('AUDIOSERVER')`

        * 'pyo' = :func:`.pyoserver.pyo_server`
        * 'jack' = :class:`.jackclient.JackClient`
        """
        if prefs.get('AUDIOSERVER') == 'pyo':
            self.server = pyoserver.pyo_server()
            self.logger.info("pyo server started")
        elif prefs.get('AUDIOSERVER') in ('jack', True):
            self.jackd = external.start_jackd()
            self.server = jackclient.JackClient()
            self.server.start()
            self.logger.info('Started jack audio server')


    def blank_LEDs(self):
        """
        If any 'LEDS' are defined in `prefs.get('HARDWARE')` ,
        instantiate them, set their color to [0,0,0],
        and then release them.
        """
        if 'LEDS' not in prefs.get('HARDWARE').keys():
            return

        for position, pins in prefs.get('HARDWARE')['LEDS'].items():
            led = gpio.LED_RGB(pins=pins)
            time.sleep(1.)
            led.set_color(col=[0,0,0])
            led.release()

    #################################################################
    # Trial Running and Management
    #################################################################
    def open_file(self):
        """
        Setup a table to store data locally.

        Opens `prefs.get('DATADIR')/local.h5`, creates a group for the current subject,
        a new table for the current day.

        .. todo::

            This needs to be unified with a general file constructor abstracted from :class:`.Subject` so it doesn't reimplement file creation!!

        Returns:
            (:class:`tables.File`, :class:`tables.Table`,
            :class:`tables.tableextension.Row`): The file, table, and row for the local data table
        """
        local_file = os.path.join(prefs.get('DATADIR'), 'local.h5')
        try:
            h5f = tables.open_file(local_file, mode='a')
        except (IOError, tables.HDF5ExtError) as e:
            self.logger.warning("local file was broken, making new")
            self.logger.warning(e)
            os.remove(local_file)
            h5f = tables.open_file(local_file, mode='w')
            os.chmod(local_file, 0o777)


        try:
            h5f.create_group("/", self.subject, "Local Data for {}".format(self.subject))
        except tables.NodeError:
            # already made it
            pass
        subject_group = h5f.get_node('/', self.subject)

        # Make a table for today's data, appending a conflict-avoidance int if one already exists
        datestring = datetime.date.today().isoformat()
        conflict_avoid = 0
        while datestring in subject_group:
            conflict_avoid += 1
            datestring = datetime.date.today().isoformat() + '-' + str(conflict_avoid)


        # Get data table descriptor
        if hasattr(self.task, 'TrialData'):
            table_descriptor = self.task.TrialData



            table = h5f.create_table(subject_group, datestring, table_descriptor,
                                               "Subject {} on {}".format(self.subject, datestring))

            # The Row object is what we write data into as it comes in
            row = table.row
            return h5f, table, row

        else:
            return h5f, None, None

    def run_task(self, task_class, task_params):
        """
        Called in a new thread, run the task.

        Opens a file with :meth:`~.autopilot.open_file` , then
        continually calls `task.stages.next` to process stages.

        Sends data back to the terminal between every stage.

        Waits for the task to clear `stage_block` between stages.
        """
        # TODO: give a net node to the Task class and let the task run itself.
        # Run as a separate thread, just keeps calling next() and shoveling data
        self.logger.debug('initializing task')
        self.task = task_class(stage_block=self.stage_block, **task_params)
        self.logger.debug('task initialized')

        # do we expect TrialData?
        trial_data = False
        if hasattr(self.task, 'TrialData'):
            trial_data = True

        # Open local file for saving
        h5f, table, row = self.open_file()

        # TODO: Init sending continuous data here
        self.logger.debug('Starting task loop')
        try:
            while True:
                # Calculate next stage data and prep triggers
                data = next(self.task.stages)() # Double parens because next just gives us the function, we still have to call it
                self.logger.debug('called stage method')

                if data:
                    data['pilot'] = self.name
                    data['subject'] = self.subject

                    # Send data back to terminal (subject is identified by the networking object)
                    self.node.send('T', 'DATA', data)

                    # Store a local copy
                    # the task class has a class variable DATA that lets us know which data the row is expecting
                    if trial_data:
                        for k, v in data.items():
                            if k in self.task.TrialData.columns.keys():
                                row[k] = v

                    # If the trial is over (either completed or bailed), flush the row
                    if 'TRIAL_END' in data.keys():
                        row.append()
                        table.flush()
                    self.logger.debug('sent data')


                # Wait on the stage lock to clear
                self.stage_block.wait()
                self.logger.debug('stage lock passed')

                # If the running flag gets set, we're closing.
                if not self.running.is_set():
                    break

        except Exception as e:
            self.logger.exception(f'got exception while running task, task stopping\n {e}')

        finally:
            self.logger.debug('stopping task')
            try:
                self.task.end()
            except Exception as e:
                self.logger.exception(f'got exception while stopping task: {e}')
            self.task = None
            row.append()
            table.flush()
            gpio.clear_scripts()
            self.logger.debug('stopped task and cleared scripts')

        h5f.flush()
        h5f.close()


if __name__ == "__main__":


    try:
        a = Pilot()
        a.quitting.wait()
    except KeyboardInterrupt:
        a.quitting.set()
        sys.exit()










import base64
import multiprocessing
import os
import socket
import threading
import time
import os
import multiprocessing
import base64
import socket
import struct
import blosc
from collections import deque
from copy import copy
from itertools import count
import typing
from typing import Optional, Union

import zmq
from tornado.ioloop import IOLoop, PeriodicCallback
from zmq.eventloop.zmqstream import ZMQStream

from autopilot import prefs
from autopilot.core.loggers import init_logger
from autopilot.networking.message import Message


class Station(multiprocessing.Process):
    """
    Independent networking class used for messaging between computers.

    These objects send and handle :class:`.networking.Message` s by using a
    dictionary of :attr:`~.networking.Station.listens`, or methods
    that are called to respond to different types of messages.

    Each sent message is given an ID, and a thread is spawned to periodically
    resend it (up until some time-to-live, typically 5 times) until confirmation
    is received.

    By default, the only listen these objects have is :meth:`.l_confirm`,
    which responds to message confirmations. Accordingly, `listens` should be added
    by using :meth:`dict.update` rather than reassigning the attribute.

    Station objects can be made with or without a :attr:`~.networking.Station.pusher`,
    a :class:`zmq.DEALER` socket that connects to the :class:`zmq.ROUTER`
    socket of an upstream Station object.

    This class can be instantiated on its own if all of the required arguments
    are supplied, but the intended pattern of use is to subclass it with
    any custom ``listen`` methods for handling message types and other logic
    that would be specific for an agent type that uses it.

    .. note::

        This object will likely be deprecated in v0.5.0, as the gains of
        a separate messaging process are not as great as the complications
        caused by having two different kinds of networking object in the system.
        In the future we will move to having a single type of networking object
        that can either be spawned as a separate process or as a thread.

    Args are similar to the documented Attributes, and so only those that differ
    from attributes are documented here

    Args:
        pusher (bool): If ``True``, create a ``zmq.DEALER`` socket connected to
            ``push_ip``, ``push_port``, and ``push_id``. (Default: ``False``).

    Attributes:
        context (:class:`zmq.Context`):  zeromq context
        loop (:class:`tornado.ioloop.IOLoop`): a tornado ioloop
        pusher (:class:`zmq.Socket`): pusher socket - a dealer socket that connects to other routers
        push_ip (str): If we have a dealer, IP to push messages to
        push_port (str):  If we have a dealer, port to push messages to
        push_id (str): :attr:`~zmq.Socket.identity` of the Router we push to
        listener (:class:`zmq.Socket`): The main router socket to send/recv messages
        listen_port (str): Port our router listens on
        logger (:class:`logging.Logger`): Used to log messages and network events.
        id (str): What are we known as? What do we set our :attr:`~zmq.Socket.identity` as?
        ip (str): Device IP
        listens (dict): Dictionary of functions to call for different types of messages. keys match the :attr:`.Message.key`.
        senders (dict): Identities of other sockets (keys, ie. directly connected) and their state (values) if they keep one
        push_outbox (dict): Messages that have been sent but have not been confirmed to our :attr:`Station.pusher`
        send_outbox (dict): Messages that have been sent but have not been confirmed to our :attr:`Station.listener`
        timers (dict): dict of :class:`threading.Timer` s that will check in on outbox messages
        msg_counter (:class:`itertools.count`): counter to index our sent messages
        file_block (:class:`threading.Event`): Event to signal when a file is being received.
    """
    repeat_interval = 5.0 # seconds to wait before retrying messages

    def __init__(self,
                 id: Optional[str] = None,
                 push_ip: Optional[str] = None,
                 push_port: Optional[int] = None,
                 push_id: Optional[str] = None,
                 pusher: bool = False,
                 listen_port: Optional[int] = None,
                 listens: Optional[typing.Dict[str, typing.Callable]] = None,
                 ):
        super(Station, self).__init__()
        # Prefs should be passed by the terminal, if not, try to load from default locatio
        # init instance attrs that are populated either by a child object or
        # in run
        self.context = None
        self.loop = None
        self.push_ip = push_ip
        self.push_port = push_port
        self.push_id = push_id.encode('utf-8') if isinstance(push_id, str) else push_id
        self.pusher = pusher # type: Union[bool, zmq.Socket]
        self.listen_port = listen_port
        self.listener = None
        self.id = id
        self.repeat_thread = None
        self.senders = {}
        self.push_outbox = {}
        self.send_outbox = {}
        self.timers = {}
        self.child = False
        self.routes = {}
        self.msgs_received = multiprocessing.Value('i', lock=True)
        self.msgs_received.value = 0

        try:
            self.ip = self.get_ip()
        except Exception as e:
            Warning("Couldn't get IP!: {}".format(e))
            self.ip = ""


        self.file_block = threading.Event() # to wait for file transfer

        # number messages as we send them
        self.msg_counter = count()

        # we have a few builtin listens
        if listens is None:
            listens = {}
        self.listens = listens
        self.listens.update({
            'CONFIRM': self.l_confirm,
            'STREAM' : self.l_stream
        })

        # even tthat signals when we are closing
        self.closing = multiprocessing.Event()
        self.closing.clear()


    def __del__(self):
        self.release()

    def run(self):
        """
        A :class:`zmq.Context` and :class:`tornado.IOLoop` are spawned,
        the listener and optionally the pusher are instantiated and
        connected to :meth:`~.Station.handle_listen` using
        :meth:`~zmq.eventloop.zmqstream.ZMQStream.on_recv` .

        The process is kept open by the :class:`tornado.IOLoop` .
        """
        try:
            self.logger = init_logger(self)
            # init zmq objects
            self.context = zmq.Context()
            self.loop = IOLoop()

            # Our networking topology is treelike:
            # each Station object binds one Router to
            # send and receive messages from its descendants
            # each Station object may have one Dealer that
            # connects it with its antecedents.
            self.listener  = self.context.socket(zmq.ROUTER)
            self.listener.setsockopt_string(zmq.IDENTITY, self.id)
            self.listener.bind('tcp://*:{}'.format(self.listen_port))
            self.listener = ZMQStream(self.listener, self.loop)
            self.listener.on_recv(self.handle_listen)

            if self.pusher is True:
                self.pusher = self.context.socket(zmq.DEALER)
                self.pusher.setsockopt_string(zmq.IDENTITY, self.id)
                self.pusher.connect('tcp://{}:{}'.format(self.push_ip, self.push_port))
                self.pusher = ZMQStream(self.pusher, self.loop)
                self.pusher.on_recv(self.handle_listen)
                # TODO: Make sure handle_listen knows how to handle ID-less messages

            # start thread that periodically resends messages
            self.repeat_thread = threading.Thread(target=self.repeat)
            self.repeat_thread.setDaemon(True)
            self.repeat_thread.start()

            self.logger.info('Starting IOLoop')
            self.loop.start()
        except KeyboardInterrupt:
            # normal quitting behavior
            pass
        finally:
            self.context.destroy()
            self.loop.close()
            # self.release()

    def prepare_message(self, to, key, value, repeat=True, flags=None):
        """
        If a message originates with us, a :class:`.Message` class
        is instantiated, given an ID and the rest of its attributes.

        Args:
            flags:
            repeat:
            to (str): The identity of the socket this message is to
            key (str): The type of message - used to select which method the receiver
                uses to process this message.
            value: Any information this message should contain. Can be any type, but
                must be JSON serializable.
        """
        msg = Message()
        msg.sender = self.id
        try:
            msg.to = to.decode('utf-8')
        except AttributeError:
            msg.to = to

        try:
            msg.key = key.decode('utf-8')
        except AttributeError:
            msg.key = key

        msg.value = value

        msg_num = next(self.msg_counter)
        msg.id = "{}_{}".format(self.id, msg_num)

        if not repeat:
            msg.flags['NOREPEAT'] = True

        if flags:
            for k, v in flags.items():
                msg.flags[k] = v

        return msg


    def send(self, to=None, key=None, value=None, msg=None, repeat=True, flags=None):
        """
        Send a message via our :attr:`~.Station.listener` , ROUTER socket.

        Either an already created :class:`.Message` should be passed as `msg`,
        or at least `to` and `key` must be provided for a new message created
        by :meth:`~.Station.prepare_message` .

        A :class:`threading.Timer` is created to resend the message using
        :meth:`~.Station.repeat` unless `repeat` is False.

        Args:
            flags:
            to (str): The identity of the socket this message is to
            key (str): The type of message - used to select which method the receiver
                uses to process this message.
            value: Any information this message should contain. Can be any type, but
                must be JSON serializable.
            msg (`.Message`): An already created message.
            repeat (bool): Should this message be resent if confirmation is not received?
        """

        if not msg and not all([to, key]):
            self.logger.exception('Need either a message or \'to\' and \'key\' fields.\
                Got\nto: {}\nkey: {}\nvalue: {}\nmsg: {}'.format(to, key, value, msg))
            return

        manual_to = False
        if not msg:
            # we're sending this ourselves, new message.
            msg = self.prepare_message(to, key, value, repeat, flags)
        elif to:
            # if given both message and to, send it to our 'to'
            # don't want to force a reserialization of the message
            manual_to = True

        if 'NOREPEAT' in msg.flags.keys():
            repeat = False

        # if we didn't send it, we shouldn't double-confirm it.
        if msg.sender not in [self.name, '_'+self.name]:
            repeat = False

        # Make sure our message has everything
        if not msg.validate():
            self.logger.exception('Message Invalid:\n{}'.format(str(msg)))

        # encode message
        msg_enc = msg.serialize()

        if not msg_enc:
            #set_trace(term_size=(80,40))
            self.logger.exception('Message could not be encoded:\n{}'.format(str(msg)))
            return

        if manual_to:
            self.listener.send_multipart([to.encode('utf-8'), msg_enc])
        else:

            if isinstance(msg.to, list):

                self.listener.send_multipart([*[hop.encode('utf-8') for hop in to], msg_enc])
            else:
                self.listener.send_multipart([msg.to.encode('utf-8'), msg_enc])

        # messages can have a flag that says not to log
        # log_this = True
        # if 'NOLOG' in msg.flags.keys():
        #     log_this = False


        #if msg.key != "CONFIRM":
        self.logger.debug('MESSAGE SENT - {}'.format(str(msg)))

        if repeat and not msg.key == "CONFIRM":
            # add to outbox and spawn timer to resend
            self.send_outbox[msg.id] = (time.time(), msg)
            # self.timers[msg.id] = threading.Timer(5.0, self.repeat, args=(msg.id,'send'))
            # self.timers[msg.id].start()
            #self.outbox.put((msg.id))

    def push(self,  to=None, key = None, value = None, msg=None, repeat=True, flags=None):
        """
        Send a message via our :attr:`~.Station.pusher` , DEALER socket.

        Unlike :meth:`~.Station.send` , `to` is not required. Every message
        is always sent to :attr:`~.Station.push_id` . `to` can be included
        to send a message further up the network tree to a networking object
        we're not directly connected to.

        Either an already created :class:`.Message` should be passed as `msg`,
        or at least `key` must be provided for a new message created
        by :meth:`~.Station.prepare_message` .

        A :class:`threading.Timer` is created to resend the message using
        :meth:`~.Station.repeat` unless `repeat` is False.

        Args:
            flags:
            to (str): The identity of the socket this message is to. If not included,
                sent to :meth:`~.Station.push_id` .
            key (str): The type of message - used to select which method the receiver
                uses to process this message.
            value: Any information this message should contain. Can be any type, but
                must be JSON serializable.
            msg (`.Message`): An already created message.
            repeat (bool): Should this message be resent if confirmation is not received?
        """
        # send message via the dealer
        # even though we only have one connection over our dealer,
        # we still include 'to' in case we are sending further upstream
        # but can push without 'to', just fill in with upstream id

        if not msg and not key:
            self.logger.exception('Need either a message or a \'key\' field.\
                Got\nto: {}\nkey: {}\nvalue: {}\nmsg: {}'.format(to, key, value, msg))

        if not msg:
            if to is None:
                to = self.push_id
            msg = self.prepare_message(to, key, value, repeat, flags)

        if 'NOREPEAT' in msg.flags.keys():
            repeat = False

        log_this = True
        if 'NOLOG' in msg.flags.keys():
            log_this = False

        # Make sure our message has everything
        if not msg.validate():
            self.logger.error('Message Invalid:\n{}'.format(str(msg)))

        # encode message
        msg_enc = msg.serialize()

        if not msg_enc:
            self.logger.error('Message could not be encoded:\n{}'.format(str(msg)))
            return

        # Even if the message is not to our upstream node, we still send it
        # upstream because presumably our target is upstream.
        self.pusher.send_multipart([self.push_id, bytes(msg.to, encoding="utf-8"), msg_enc])

        if not (msg.key == "CONFIRM") and log_this:
            self.logger.debug('MESSAGE PUSHED - {}'.format(str(msg)))

        if repeat and not msg.key == 'CONFIRM':
            # add to outbox and spawn timer to resend
            self.push_outbox[msg.id] = (time.time(), msg)
            #self.timers[msg.id] = threading.Timer(5.0, self.repeat, args=(msg.id, 'push'))
            #self.timers[msg.id].start()

    def repeat(self):
        """
        Periodically (according to :attr:`~.repeat_interval`) resend messages that haven't been confirmed

        TTL is decremented, and messages are resent until their TTL is 0.

        """
        while not self.closing.is_set():
            # make local copies
            push_outbox = copy(self.push_outbox)
            send_outbox = copy(self.send_outbox)

            # try to send any outstanding messages and delete if too old
            if len(push_outbox)>0:
                for id in push_outbox.keys():
                    if push_outbox[id][1].ttl <= 0:
                        self.logger.warning('PUBLISH FAILED {} - {}'.format(id, str(push_outbox[id][1])))
                        try:
                            del self.push_outbox[id]
                        except KeyError:
                            # fine, already deleted
                            pass
                    else:
                        # if we didn't just put this message in our outbox
                        if (time.time() - push_outbox[id][0]) > self.repeat_interval*2:

                            self.logger.debug('REPUBLISH {} - {}'.format(id, str(push_outbox[id][1])))
                            self.pusher.send_multipart([self.push_id, push_outbox[id][1].serialize()])
                            self.push_outbox[id][1].ttl -= 1



            if len(send_outbox)>0:
                for id in send_outbox.keys():
                    if send_outbox[id][1].ttl <= 0:
                        self.logger.warning('PUBLISH FAILED {} - {}'.format(id, str(send_outbox[id][1])))
                        try:
                            del self.send_outbox[id]
                        except KeyError:
                            # fine, already deleted
                            pass

                    else:
                        # if we didn't just put this message in our outbox
                        if (time.time() - send_outbox[id][0]) > self.repeat_interval*2:

                            self.logger.debug('REPUBLISH {} - {}'.format(id, str(send_outbox[id][1])))
                            self.listener.send_multipart([bytes(send_outbox[id][1].to, encoding="utf-8"), send_outbox[id][1].serialize()])
                            self.send_outbox[id][1].ttl -= 1

            # wait to do it again
            time.sleep(self.repeat_interval)



    def l_confirm(self, msg):
        """
        Confirm that a message was received.

        Args:
            msg (:class:`.Message`): A confirmation message - note that this message has its own unique ID, so the value of this message contains the ID of the message that is being confirmed
        """
        # confirmation that a published message was received
        # value should be the message id

        # delete message from outbox if we still have it
        try:
            if msg.value in self.send_outbox.keys():
                del self.send_outbox[msg.value]
            elif msg.value in self.push_outbox.keys():
                del self.push_outbox[msg.value]
        except KeyError:
            # fine, already deleted
            pass

        # if this is a message to our internal net_node, make sure it gets the memo that shit was confirmed too
        if msg.to == "_{}".format(self.id):
            self.send("_{}".format(self.id), 'CONFIRM', msg.value)


        #self.logger.info('CONFIRMED MESSAGE {}'.format(msg.value))

    def l_stream(self, msg):
        """
        Reconstitute the original stream of messages and call their handling methods

        The ``msg`` should contain an ``inner_key`` that indicates the key, and thus the
        handling method.

        Args:
            msg (dict): Compressed stream sent by :meth:`Net_Node._stream`
        """
        listen_fn = self.listens[msg.value['inner_key']]
        old_value = copy(msg.value)
        delattr(msg, 'value')
        msg.key = old_value['inner_key']
        for v in old_value['payload']:
            if isinstance(v, dict) and ('headers' in old_value.keys()):
                v.update(old_value['headers'])
            msg.value = v
            listen_fn(msg)

    def handle_listen(self, msg:typing.List[bytes]):
        """
        Upon receiving a message, call the appropriate listen method
        in a new thread.

        If the message is :attr:`~.Message.to` us, send confirmation.

        If the message is not :attr:`~.Message.to` us, attempt to forward it.

        Args:
            msg (str): JSON :meth:`.Message.serialize` d message.
        """
        # TODO: This check is v. fragile, pyzmq has a way of sending the stream along with the message
        #####################33
        # Parse the message
        with self.msgs_received.get_lock():
            self.msgs_received.value += 1

        if msg[-1] == b'CLOSING':
            self.loop.stop()

        if len(msg)==1:
            # from our dealer, these are always to us.
            send_type = 'dealer'
            #msg = json.loads(msg[0])
            #msg = Message(**msg)
            msg = Message(msg[0])

        elif len(msg)>=2:
            # from the router
            send_type = 'router'
            sender = msg[0]

            # if this message was a multihop message, store the route
            if len(msg)>4:
                self.routes[sender] = msg[0:-3]

            # # if this is a new sender, add them to the list
            if sender not in self.senders.keys():
                self.senders[sender] = ""
                self.senders[b'_' + sender] = ''

            # connection pings are blank frames,
            # respond to let them know we're alive
            if msg[-1] == b'':
                self.listener.send_multipart(msg)
                return

            # if this message wasn't to us, forward without deserializing
            # the second to last should always be the intended recipient
            unserialized_to = msg[-2]
            if unserialized_to.decode('utf-8') not in [self.id, "_{}".format(self.id)]:
                # forward it!!
                if len(msg) > 4:
                    # multihop message, just determine whether the next hop is through
                    # our pusher or router
                    if self.pusher and msg[2] not in self.senders.keys():
                        self.pusher.send_multipart(msg[2:])
                        self.logger.debug(f'FORWARDING (multihop dealer): {msg}')
                    else:
                        self.listener.send_multipart(msg[2:])
                        self.logger.debug(f'FORWARDING (multihop router): {msg}')
                else:
                    if unserialized_to not in self.senders.keys() and self.pusher:
                        # if we don't know who they are and we have a pusher, try to push it
                        self.pusher.send_multipart([self.push_id, *msg[2:]])
                        self.logger.debug(f'FORWARDING (dealer): {msg}')
                    else:
                        # if we know who they are or not, try to send it through router anyway.
                        # send everything but the first two frames, which should be the ID of
                        # the sender and us
                        self.listener.send_multipart(msg[2:])
                        self.logger.debug(f'FORWARDING (router): {msg}')

                return

            msg = Message(msg[-1])

            # if this is a new sender, add them to the list
            if msg['sender'] not in self.senders.keys():
                self.senders[msg['sender']] = ""
                self.senders['_' + msg['sender']] = ''

        else:
            self.logger.error('Dont know what this message is:{}'.format(msg))
            return

        ###################################
        # Handle the message

        # unstack and listed to fields remaining from forwarding
        if isinstance(msg.to, list):
            if len(msg.to) == 1:
                msg.to = msg.to[0]

        # if this message is to us, just handle it and return
        if msg.to in [self.id, "_{}".format(self.id)]:
            if (msg.key != "CONFIRM"):
                self.logger.debug('RECEIVED: {}'.format(str(msg)))
            # Log and spawn thread to respond to listen
            try:
                listen_funk = self.listens[msg.key]
                listen_thread = threading.Thread(target=listen_funk, args=(msg,))
                listen_thread.start()
            except KeyError:
                self.logger.exception('No function could be found for msg id {} with key: {}'.format(msg.id, msg.key))


            # send a return message that confirms even if we except
            # don't confirm confirmations
            if (msg.key != "CONFIRM") and ('NOREPEAT' not in msg.flags.keys()):
                if send_type == 'router':
                    self.send(msg.sender, 'CONFIRM', msg.id)
                elif send_type == 'dealer':
                    self.push(msg.sender, 'CONFIRM', msg.id)
            return
        elif self.child and (msg.to == 'T'):
            # FIXME UGLY HACK
            self.push(msg=msg)
        else:
            self.logger.exception(f'Message not to us, but wasnt forwarded previously in handling method, message must be misformatted: {msg}')

        # all of the below should be removed but i am paranoid and like
        # sorta bad at reverting changes on git so i am just commenting for now

        #
        #
        # elif msg.to in self.senders.keys():
        #     self.send(msg=msg)
        # # otherwise, if we have a pusher, send it there
        # # it's either for them or some other upstream node we don't know about
        # elif self.pusher:
        #     self.push(msg=msg)
        # else:
        #
        #     self.logger.warning('Message to unconfirmed recipient, attempting to send: {}'.format(str(msg)))
        #     self.send(msg=msg)

        # finally, if there's something we're supposed to do, do it
        # even if the message is not to us,
        # sometimes we do work en passant to reduce effort doubling
        # FIXME Seems like a really bad idea.
        # if msg.key in self.listens.keys():
        #     listen_funk = self.listens[msg.key]
        #     listen_thread = threading.Thread(target=listen_funk, args=(msg,))
        #     listen_thread.start()
        #
        # # since we return if it's to us before, confirm is repeated down here.
        # # FIXME: Inelegant
        # if (msg.key != "CONFIRM") and ('NOREPEAT' not in msg.flags.keys()):
        #     if send_type == 'router':
        #         self.send(sender, 'CONFIRM', msg.id)
        #     elif send_type == 'dealer':
        #         self.push(msg.sender, 'CONFIRM', msg.id)

    def get_ip(self):
        """
        Find our IP address

        returns (str): our IPv4 address.
        """

        # shamelessly stolen from https://www.w3resource.com/python-exercises/python-basic-exercise-55.php
        # variables are badly named because this is just a rough unwrapping of what was a monstrous one-liner
        # (and i don't really understand how it works)

        # get ips that aren't the loopback
        unwrap00 = [ip for ip in socket.gethostbyname_ex(socket.gethostname())[2] if not ip.startswith("127.")][:1]
        # ??? truly dk
        unwrap01 = [[(s.connect(('8.8.8.8', 53)), s.getsockname()[0], s.close()) for s in
                     [socket.socket(socket.AF_INET, socket.SOCK_DGRAM)]][0][1]]

        unwrap2 = [l for l in (unwrap00, unwrap01) if l][0][0]

        return unwrap2

    def release(self):

        self.closing.set()

        # send a message to ourselves from the parent process to this one
        ctx = zmq.Context().instance()
        sock = ctx.socket(zmq.DEALER)
        sock.setsockopt_string(zmq.IDENTITY, f"{self.id}.closer")
        sock.connect(f'tcp://localhost:{self.listen_port}')
        sock.send_multipart([self.id.encode('utf-8'), 'CLOSING'.encode('utf-8')])
        sock.close()

        # self.loop.stop()
        self.terminate()
        # except AttributeError:
        #     # already been called, NoneTypes have no attribute terminate
        #     pass

        # Stopping the loop should kill the process, as it's what's holding us in run()
        #self.loop.stop()

    def _check_stop(self):
        """
        periodic callback called by the IOLoop to check if the `closing` flag has been set, and closing process if so
        """
        if self.closing.is_set():
            self.loop.stop()


class Terminal_Station(Station):
    """
    :class:`~.networking.Station` object used by :class:`~.Terminal`
    objects.

    Spawned without a :attr:`~.Station.pusher`.

    **Listens**

    +-------------+-------------------------------------------+-----------------------------------------------+
    | Key         | Method                                    | Description                                   |
    +=============+===========================================+===============================================+
    | 'PING'      | :meth:`~.Terminal_Station.l_ping`         | We are asked to confirm that we are alive     |
    +-------------+-------------------------------------------+-----------------------------------------------+
    | 'INIT'      | :meth:`~.Terminal_Station.l_init`         | Ask all pilots to confirm that they are alive |
    +-------------+-------------------------------------------+-----------------------------------------------+
    | 'CHANGE'    | :meth:`~.Terminal_Station.l_change`       | Change a parameter on the Pi                  |
    +-------------+-------------------------------------------+-----------------------------------------------+
    | 'STOPALL'   | :meth:`~.Terminal_Station.l_stopall`      | Stop all pilots and plots                     |
    +-------------+-------------------------------------------+-----------------------------------------------+
    | 'KILL'      | :meth:`~.Terminal_Station.l_kill`         | Terminal wants us to die :(                   |
    +-------------+-------------------------------------------+-----------------------------------------------+
    | 'DATA'      | :meth:`~.Terminal_Station.l_data`         | Stash incoming data from a Pilot              |
    +-------------+-------------------------------------------+-----------------------------------------------+
    | 'STATE'     | :meth:`~.Terminal_Station.l_state`        | A Pilot has changed state                     |
    +-------------+-------------------------------------------+-----------------------------------------------+
    | 'HANDSHAKE' | :meth:`~.Terminal_Station.l_handshake`    | A Pi is telling us it's alive and its IP      |
    +-------------+-------------------------------------------+-----------------------------------------------+
    | 'FILE'      | :meth:`~.Terminal_Station.l_file`         | The pi needs some file from us                |
    +-------------+-------------------------------------------+-----------------------------------------------+

    """

    plot_timer = None
    #send_plot = threading.Event()
    #send_plot.clear()
    # dict of threading events that determine how frequently we send plot updates
    sent_plot = {}

    def __init__(self, pilots):
        """
        Args:
            pilots (dict): The :attr:`.Terminal.pilots` dictionary.
        """
        super(Terminal_Station, self).__init__()

        # by default terminal doesn't have a pusher, everything connects to it
        self.pusher = False

        # Store some prefs values
        self.listen_port = prefs.get('MSGPORT')
        self.id = 'T'

        # Message dictionary - What method to call for each type of message received by the terminal class
        self.listens.update({
            'PING':      self.l_ping,  # We are asked to confirm that we are alive
            'INIT':      self.l_init,  # We should ask all the pilots to confirm that they are alive
            'CHANGE':    self.l_change,  # Change a parameter on the Pi
            'STOPALL':   self.l_stopall, # Stop all pilots and plots
            'KILL':      self.l_kill,  # Terminal wants us to die :(
            'DATA':      self.l_data,  # Stash incoming data from an autopilot
            'CONTINUOUS': self.l_continuous, # handle incoming continuous data
            'STATE':     self.l_state,  # The Pi is confirming/notifying us that it has changed state
            'HANDSHAKE': self.l_handshake, # initial connection with some initial info
            'FILE':      self.l_file,  # The pi needs some file from us
        })

        # dictionary that keeps track of our pilots
        self.pilots = pilots

        # start a timer at the draw FPS of the terminal -- only send
        if prefs.get( 'DRAWFPS'):
            self.data_fps = float(prefs.get('DRAWFPS'))
        else:
            self.data_fps = 20
        self.data_ifps = 1.0/self.data_fps



    def start_plot_timer(self):
        """
        Start a timer that controls how often streamed video frames are sent to
        :class:`.gui.Video` plots.
        """
        self.plot_timer = threading.Thread(target=self._fps_clock)
        self.plot_timer.setDaemon(True)
        self.plot_timer.start()


    def _fps_clock(self):
        while not self.closing.is_set():
            for k, v in self.sent_plot.items():
                try:
                    v.set()
                except:
                    pass
                    # TODO: Too General

            #self.send_plot.set()
            time.sleep(self.data_ifps)


    ##########################
    # Message Handling Methods

    def l_ping(self, msg:Message):
        """
        We are asked to confirm that we are alive

        Respond with a blank 'STATE' message.

        Args:
            msg (:class:`.Message`):
        """
        # we are being asked if we're alive
        # respond with blank message since the terminal isn't really stateful
        self.send(msg.sender, 'STATE', flags={'NOLOG':True})

    def l_init(self, msg:Message):
        """
        Ask all pilots to confirm that they are alive

        Sends a "PING" to everyone in the pilots dictionary.

        Args:
            msg (:class:`.Message`):
        """
        # Ping all pis that we are expecting given our pilot db
        # Responses will be handled with l_state so not much needed here

        for p in self.pilots.keys():
            self.send(p, 'PING', flags={'NOLOG':True})


    def l_change(self, msg:Message):
        """
        Change a parameter on the Pi

        Warning:
            Not Implemented

        Args:
            msg (:class:`.Message`):
        """
        # TODO: Should also handle param changes to GUI objects like ntrials, etc.
        pass

    def l_stopall(self, msg:Message):
        """
        Stop all pilots and plots

        Args:
            msg (:class:`.Message`):
        """
        # let all the pilots and plot objects know that they should stop
        for p in self.pilots.keys():
            self.send(p, 'STOP')
            self.send("P_{}".format(p), 'STOP')


    def l_kill(self, msg:Message):
        """
        Terminal wants us to die :(

        Stop the :attr:`.Station.loop`

        Args:
            msg (:class:`.Message`):
        """
        self.logger.info('Received kill request')

        self.closing.set()

        # Stopping the loop should kill the process, as it's what's holding us in run()
        self.loop.stop()


    def l_data(self, msg:Message):
        """
        Stash incoming data from a Pilot

        Just forward this along to the internal terminal object ('_T')
        and a copy to the relevant plot.

        Args:
            msg (:class:`.Message`):
        """
        # Send through to terminal
        #self.send('_T', 'DATA', msg.value, flags=msg.flags)
        self.send(to='_T', msg=msg)

        # Send to plot widget, which should be listening to "P_{pilot_name}"
        #self.send('P_{}'.format(msg.value['pilot']), 'DATA', msg.value, flags=msg.flags)
        self.send(to='P_{}'.format(msg.value['pilot']), msg=msg)

    def l_continuous(self, msg:Message):
        """
        Handle the storage of continuous data

        Forwards all data on to the Terminal's internal :class:`Net_Node`,
        send to :class:`.Plot` according to update rate in ``prefs.get('DRAWFPS')``

        Args:
            msg (:class:`.Message`): A continuous data message
        """



        # Send through to terminal
        #msg.value.update({'continuous':True})
        self.send(to='_T', msg=msg)

        # Send to plot widget, which should be listening to "P_{pilot_name}"
        plot_id = 'P_{}'.format(msg.value['pilot'])
        if plot_id in self.senders.keys():
            if not self.plot_timer:
                self.start_plot_timer()

            if msg.sender not in self.sent_plot.keys():
                self.sent_plot[msg.sender] = threading.Event()
            if self.sent_plot[msg.sender].is_set():
                self.send(to=plot_id, msg=msg)
                self.sent_plot[msg.sender].clear()


    # def l_continuous(self, msg):
    #
    #     # Send through to terminal
    #     msg.value.update({'continuous':True})
    #     self.send('_T', 'DATA', msg.value, flags=msg.flags)
    #
    #     # Send to plot widget, which should be listening to "P_{pilot_name}"
    #     self.send('P_{}'.format(msg.value['pilot']), 'DATA', msg.value, flags=msg.flags)



    def l_state(self, msg:Message):
        """
        A Pilot has changed state.

        Stash in 'state' field of pilot dict and send along to _T

        Args:
            msg (:class:`.Message`):
        """
        if msg.sender not in self.pilots.keys():
            self.pilots[msg.sender] = {}
            #if 'state' in self.pilots[msg.sender].keys():
                # if msg.value == self.pilots[msg.sender]['state']:
                #     # if we've already gotten this one, don't send to terminal
                #     return

        self.pilots[msg.sender]['state'] = msg.value

        # Tell the terminal so it can update the pilot_db file
        state = {'state':msg.value, 'pilot':msg.sender}
        self.send('_T', 'STATE', state)

        # Tell the plot
        self.send("P_{}".format(msg.sender), 'STATE', msg.value)

        self.senders[msg.sender] = msg.value

    def l_handshake(self, msg:Message):
        """
        A Pi is telling us it's alive and its IP.

        Send along to _T

        Args:
            msg (:class:`.Message`):
        """
        # only rly useful for our terminal object
        self.send('_T', 'HANDSHAKE', value=msg.value)



    def l_file(self, msg:Message):
        """
        A Pilot needs some file from us.

        Send it back after :meth:`base64.b64encode` ing it.

        TODO:
            Split large files into multiple messages...

        Args:
            msg (:class:`.Message`): The value field of the message should contain some
                relative path to a file contained within `prefs.get('SOUNDDIR')` . eg.
                `'/songs/sadone.wav'` would return `'os.path.join(prefs.get('SOUNDDIR')/songs.sadone.wav'`
        """
        # The <target> pi has requested some file <value> from us, let's send it back
        # This assumes the file is small, if this starts crashing we'll have to split the message...

        full_path = os.path.join(prefs.get('SOUNDDIR'), msg.value)
        with open(full_path, 'rb') as open_file:
            # encode in base64 so json doesn't complain
            file_contents = base64.b64encode(open_file.read())

        file_message = {'path':msg.value, 'file':file_contents}

        self.send(msg.sender, 'FILE', file_message)


class Pilot_Station(Station):
    """
    :class:`~.networking.Station` object used by :class:`~.Pilot`
    objects.

    Spawned with a :attr:`~.Station.pusher` connected back to the
    :class:`~.Terminal` .

    **Listens**

    +-------------+-------------------------------------+-----------------------------------------------+
    | Key         | Method                              | Description                                   |
    +=============+=====================================+===============================================+
    | 'STATE'     | :meth:`~.Pilot_Station.l_state`     | Pilot has changed state                       |
    | 'COHERE'    | :meth:`~.Pilot_Station.l_cohere`    | Make sure our data and the Terminal's match.  |
    | 'PING'      | :meth:`~.Pilot_Station.l_ping`      | The Terminal wants to know if we're listening |
    | 'START'     | :meth:`~.Pilot_Station.l_start`     | We are being sent a task to start             |
    | 'STOP'      | :meth:`~.Pilot_Station.l_stop`      | We are being told to stop the current task    |
    | 'PARAM'     | :meth:`~.Pilot_Station.l_change`    | The Terminal is changing some task parameter  |
    | 'FILE'      | :meth:`~.Pilot_Station.l_file`      | We are receiving a file                       |
    +-------------+-------------------------------------+-----------------------------------------------+

    """
    def __init__(self):
        # Pilot has a pusher - connects back to terminal
        super(Pilot_Station, self).__init__()

        self.pusher = True
        if prefs.get('LINEAGE') == 'CHILD':
            self.push_id = prefs.get('PARENTID').encode('utf-8')
            self.push_port = prefs.get('PARENTPORT')
            self.push_ip = prefs.get('PARENTIP')
            self.child = True

        else:
            self.push_id = b'T'
            self.push_port = prefs.get('PUSHPORT')
            self.push_ip = prefs.get('TERMINALIP')
            self.child = False

        # Store some prefs values
        self.listen_port = prefs.get('MSGPORT')

        #self.id = prefs.get('NAME').encode('utf-8')
        self.id = prefs.get('NAME')
        self.pi_id = "_{}".format(self.id)
        self.subject = None # Store current subject ID
        self.state = 'IDLE' # store current pi state
        self.child = False # Are we acting as a child right now?
        self.parent = False # Are we acting as a parent right now?


        self.listens.update({
            'STATE': self.l_state,  # Confirm or notify terminal of state change
            'COHERE': self.l_cohere, # Sending our temporary data table at the end of a run to compare w/ terminal's copy
            'PING': self.l_ping,  # The Terminal wants to know if we're listening
            'START': self.l_start,  # We are being sent a task to start
            'STOP': self.l_stop,  # We are being told to stop the current task
            'PARAM': self.l_change,  # The Terminal is changing some task parameter
            'FILE': self.l_file,  # We are receiving a file
            'CONTINUOUS': self.l_continuous, # we are sending continuous data to the terminal
            'CHILD': self.l_child,
            'HANDSHAKE': self.l_noop,
            'CALIBRATE_PORT': self.l_forward,
            'CALIBRATE_RESULT': self.l_forward,
            'BANDWIDTH': self.l_forward,
            'STREAM_VIDEO': self.l_forward
        })

        # ping back our status to the terminal every so often
        if prefs.get('PING_INTERVAL') is None:
            self.ping_interval = 5
        else:
            self.ping_interval = float(prefs.get('PING_INTERVAL'))

        self._ping_thread = threading.Timer(self.ping_interval, self._pinger)
        self._ping_thread.setDaemon(True)
        self._ping_thread.start()

    def _pinger(self):
        """
        Periodically ping the terminal with our status

        Calls its own timer to replace it


        Returns:

        """
        # before .run is called, pusher is a boolean flag telling us to make one when it is
        # not great and should be changed, but these modules are not long for this world
        # in their current form anyway
        if not isinstance(self.pusher, bool):
            self.l_ping()
        if not self.closing.is_set():
            self._ping_thread = threading.Timer(self.ping_interval, self._pinger)
            self._ping_thread.setDaemon(True)
            self._ping_thread.start()


    ###########################3
    # Message/Listen handling methods

    def l_noop(self, msg):
        pass

    def l_state(self, msg:Message):
        """
        Pilot has changed state

        Stash it and alert the Terminal

        Args:
            msg (:class:`.Message`):
        """
        # Save locally so we can respond to queries on our own, then push 'er on through
        # Value will just have the state, we want to add our name
        self.state = msg.value

        self.push(to=self.push_id, key='STATE', value=msg.value)


    def l_cohere(self, msg:Message):
        """
        Send our local version of the data table so the terminal can double check

        Warning:
            Not Implemented

        Args:
            msg (:class:`.Message`):
        """

        pass

    def l_ping(self, msg:Message=None):
        """
        The Terminal wants to know our status

        Push back our current state.

        Args:
            msg (:class:`.Message`):
        """
        # The terminal wants to know if we are alive, respond with our name and IP
        # don't bother the pi
        self.push(key='STATE', value=self.state, flags={'NOLOG':True})

    def l_start(self, msg:Message):
        """
        We are being sent a task to start

        If we need any files, request them.

        Then send along to the pilot.

        Args:
            msg (:class:`.Message`): value will contain a dictionary containing a task
                description.
        """
        self.subject = msg.value['subject']

        # TODO: Refactor into a general preflight check.
        # First make sure we have any sound files that we need
        # TODO: stim managers need to be able to return list of stimuli and this is a prime reason why
        if 'stim' in msg.value.keys():
            if 'sounds' in msg.value['stim'].keys():

                # nested list comprehension to get value['sounds']['L/R'][0-n]
                f_sounds = [sound for sounds in msg.value['stim']['sounds'].values() for sound in sounds
                            if sound['type'] in ['File', 'Speech']]
            elif 'manager' in msg.value['stim'].keys():
                # we have a manager
                if msg.value['stim']['type'] == 'sounds':
                    f_sounds = []
                    for group in msg.value['stim']['groups']:
                        f_sounds.extend([sound for sounds in group['sounds'].values() for sound in sounds
                                        if sound['type'] in ['File', 'Speech']])
            else:
                f_sounds = []

            if len(f_sounds)>0:
                # check to see if we have these files, if not, request them
                for sound in f_sounds:
                    full_path = os.path.join(prefs.get('SOUNDDIR'), sound['path'])
                    if not os.path.exists(full_path):
                        # We ask the terminal to send us the file and then wait.
                        self.logger.info('REQUESTING SOUND {}'.format(sound['path']))
                        self.push(key='FILE', value=sound['path'])
                        # wait here to get the sound,
                        # the receiving thread will set() when we get it.
                        self.file_block.clear()
                        self.file_block.wait()

        # If we're starting the task as a child, stash relevant params
        if 'child' in msg.value.keys():
            self.child = True
            self.parent_id = msg.value['child']['parent']
            self.subject = msg.value['child']['subject']

        else:
            self.child = False


        # once we make sure we have everything, tell the Pilot to start.
        self.send(self.pi_id, 'START', msg.value)

    def l_stop(self, msg:Message):
        """
        Tell the pi to stop the task

        Args:
            msg (:class:`.Message`):
        """
        self.send(self.pi_id, 'STOP')

    def l_change(self, msg:Message):
        """
        The terminal is changing a parameter

        Warning:
            Not implemented

        Args:
            msg (:class:`.Message`):
        """
        # TODO: Changing some task parameter from the Terminal
        pass

    def l_file(self, msg:Message):
        """
        We are receiving a file.

        Decode from b64 and save. Set the file_block.

        Args:
            msg (:class:`.Message`): value will have 'path' and 'file',
                where the path determines where in `prefs.get('SOUNDDIR')` the
                b64 encoded 'file' will be saved.
        """
        # The file should be of the structure {'path':path, 'file':contents}

        full_path = os.path.join(prefs.get('SOUNDDIR'), msg.value['path'])
        # TODO: give Message full deserialization capabilities including this one
        file_data = base64.b64decode(msg.value['file'])
        try:
            os.makedirs(os.path.dirname(full_path))
        except:
            # TODO: Make more specific - only if dir already exists
            pass
        with open(full_path, 'wb') as open_file:
            open_file.write(file_data)

        self.logger.info('SOUND RECEIVED {}'.format(msg.value['path']))

        # If we requested a file, some poor start fn is probably waiting on us
        self.file_block.set()

    def l_continuous(self, msg:Message):
        """
        Forwards continuous data sent by children back to terminal.

        Continuous data sources from this pilot should be streamed directly to the terminal.

        Args:
            msg (:class:`Message`): Continuous data message

        """
        if self.child:
            msg.value['pilot'] = self.parent_id
            msg.value['subject'] = self.subject
            msg.value['continuous'] = True
            self.push(to='T', key='DATA', value=msg.value, repeat=False)
        else:
            self.logger.warning('Received continuous data but no child found, \
                                continuous data should be streamed directly to terminal \
                                from pilot')

    def l_child(self, msg:Message):
        """
        Tell one or more children to start running a task.

        By default, the `key` argument passed to `self.send` is 'START'.
        However, this can be overriden by providing the desired string
        as `msg.value['KEY']`.

        This checks the pref `CHILDID` to get the names of one or more children.
        If that pref is a string, sends the message to just that child.
        If that pref is a list, sends the message to each child in the list.

        Args:
            msg (): A message to send to the child or children.

        Returns: 
            nothing
        """
        # Take `KEY` from msg.value['KEY'] if available
        # Otherwise, use 'START'
        if 'KEY' in msg.value.keys():
            KEY = msg.value['KEY']
        else:
            KEY = 'START'
        
        # Get the list of children
        pref_childid = prefs.get('CHILDID')
        
        # Send to one or more children
        if isinstance(pref_childid, list):
            # It's a list of multiple children, send to each
            for childid in pref_childid:
                self.send(to=childid, key=KEY, value=msg.value)
        else:
            # Send to the only child
            self.send(to=pref_childid, key=KEY, value=msg.value)

    def l_forward(self, msg:Message):
        """
        Just forward the message to the pi.
        """
        self.send(to=self.pi_id, key=msg.key, value=msg.value)


from autopilot import prefs

from autopilot.stim.stim import Stim
from autopilot.stim.managers import Stim_Manager, Proportional, init_manager


if prefs.get('AGENT') == "pilot":
    if 'AUDIO' in prefs.get('CONFIG'):
        from autopilot.stim.sound import sounds





HAVE_PYSIDE = False
try:
    from PySide2 import QtCore
    HAVE_PYSIDE = True
except ImportError:
    pass

_INVOKER = None

if HAVE_PYSIDE:

    class InvokeEvent(QtCore.QEvent):
        """
        Sends signals to the main QT thread from spawned message threads

        See `stackoverflow <https://stackoverflow.com/a/12127115>`_
        """

        EVENT_TYPE = QtCore.QEvent.Type(QtCore.QEvent.registerEventType())

        def __init__(self, fn, *args, **kwargs):
            """
            Accepts a function, its args and kwargs and wraps them as a
            :class:`QtCore.QEvent`

            """
            QtCore.QEvent.__init__(self, InvokeEvent.EVENT_TYPE)
            self.fn = fn
            self.args = args
            self.kwargs = kwargs


    class Invoker(QtCore.QObject):
        """
        Wrapper that calls an evoked event made by :class:`.InvokeEvent`
        """
        def event(self, event):
            """
            Args:
                event:
            """
            event.fn(*event.args, **event.kwargs)
            return True


def get_invoker():
    if not globals()['HAVE_PYSIDE']:
        raise Exception("PySide2 could not be imported, no GUI event invoker can be gotten")
    if globals()['_INVOKER'] is None:
        globals()['_INVOKER'] = Invoker()
    return globals()['_INVOKER']


import requests
import sys
import re
import json
import os
from tqdm import tqdm

def download_box(file_url, filename = None, save_path = None):
    """
    Download a file from Box from a URL.

    eg. https://flir.app.boxcn.net/v/SpinnakerSDK/file/545648953427

    Note::

        Only works for files that don't require authentication

    Args:
        file_url (str): A URL of a preview page on Box
        filename (str): Filename (without path) to save file. If None, the filename from the server is used
        save_path (str): A path to save the file to. If None, current directory is used

    Returns:
        filename (str): the complete path that the file has been saved to

    """
    sess = requests.Session()


    if isinstance(file_url, bytes):
        file_url = file_url.decode('utf-8')

    if isinstance(filename, bytes):
        filename = filename.decode('utf-8')

    if isinstance(save_path, bytes):
        save_path = save_path.decode('utf-8')


    file_name = "file_"+file_url.split('/')[-1]
    file_number = file_url.split('/')[-1]
    base_url = '/'.join(file_url.split('/')[0:3])
    share_url = '/'.join(file_url.split('/')[0:5])

    page = sess.get(file_url,
        headers={
            "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8",
            "Accept-Language": "en-US,en;q=0.5",
            "Connection": "keep-alive",
            "DNT": "1",
            "Upgrade-Insecure-Requests": "1",
            "User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10.15; rv:70.0) Gecko/20100101 Firefox/70.0"
        },
        cookies={},
    )



    req_parts = {
        'z': page.cookies.get('z'),
        'box_visitor_id': page.cookies.get('box_visitor_id'),
        'bv': page.cookies.get('bv'),
        'cn': page.cookies.get('cn'),
        'token': re.search('(?<=\"requestToken\":\")(.+?)(?=\")', page.content.decode('utf-8')).group(0)
    }


    tokens = sess.post("https://flir.app.boxcn.net/app-api/enduserapp/elements/tokens",
        json={'fileIDs':[file_name]},
        headers={
            "Accept": "application/json",
            "Accept-Language": "en-US,en;q=0.5",
            "Connection": "keep-alive",
            "Content-Type": "application/json",
            "DNT": "1",
            "Origin": base_url,
            "Referer": file_url,
            "Request-Token": "{}".format(req_parts['token']),
            "User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10.15; rv:70.0) Gecko/20100101 Firefox/70.0",
            "X-Box-Client-Name": "enduserapp",
            "X-Box-Client-Version": "20.162.1",
            "X-Box-EndUser-API": "vanityName=SpinnakerSDK",
            "X-Request-Token": "{}".format(req_parts['token'])
        },
    )

    req_parts['dl_token'] = json.loads(tokens.content)[file_name]['read']

    dl_link = sess.get("https://api.box.com/2.0/files/{}?fields=download_url".format(file_number),
        headers={
            "Accept": "application/json, text/plain, */*",
            "Accept-Language": "en-US,en;q=0.5",
            "Authorization": "Bearer {}".format(req_parts['dl_token']),
            "BoxApi": "shared_link={}".format(share_url),
            "Connection": "keep-alive",
            "DNT": "1",
            "Origin": base_url,
            "Referer": file_url,
            "User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10.15; rv:70.0) Gecko/20100101 Firefox/70.0",
            "X-Box-Client-Name": "box-content-preview",
            "X-Box-Client-Version": "2.23.0",
            "X-Rep-Hints": "[3d][pdf][text][mp3][jpg?dimensions=1024x1024&paged=false][jpg?dimensions=2048x2048,png?dimensions=2048x2048][dash,mp4][filmstrip]"
        },
    )

    dl_url = json.loads(dl_link.content)['download_url']

    dl_file = sess.get(dl_url, stream=True)

    if filename is None:
        filename = re.search("filename=\"(.+?)\";", dl_file.headers.get('Content-Disposition')).groups(0)[0]
    if save_path:
        filename = os.path.join(save_path, filename)

    total_size = int(dl_file.headers.get('content-length', 0))
    block_size = 1024  # 1 Kibibyte
    t = tqdm(total=total_size, unit='iB', unit_scale=True)
    with open(filename, 'wb') as output_file:
        for data in dl_file.iter_content(block_size):
            t.update(len(data))
            output_file.write(data)
    t.close()

    sys.stdout.write(filename)



"""
Data transformations.

Experimental module.

Reusable transformations from one representation of data to another.
eg. converting frames of a video to locations of objects,
or locations of objects to area labels

.. todo::

    This is a preliminary module and it purely synchronous at the moment. It will be expanded to ...
    * support multiple asynchronous processing rhythms
    * support automatic value coercion

    The following design features need to be added
    * recursion checks -- make sure a child hasn't already been added to a processing chain.
"""

import types
import typing
from enum import Enum, auto
from autopilot.core.loggers import init_logger

class TransformRhythm(Enum):
    """
    Attributes:
        FIFO: First-in-first-out, process inputs as they are received, potentially slowing down the transformation pipeline
        FILO: First-in-last-out, process the most recent input, ignoring previous (lossy transformation)
    """
    FIFO = auto()
    FILO = auto()

class Transform(object):
    """
    Metaclass for data transformations

    Each subclass should define the following

    * :meth:`.process` - a method that takes the input of the transoformation as its single argument and returns the transformed output
    * :attr:`.format_in` - a `dict` that specifies the input format
    * :attr:`.format_out` - a `dict` that specifies the output format

    Arguments:
        rhythm (:class:`TransformRhythm`): A rhythm by which the transformation object processes its inputs

    Attributes:
        child (class:`Transform`): Another Transform object chained after this one
    """

    def __init__(self, rhythm : TransformRhythm = TransformRhythm.FILO, *args, **kwargs):
        self._child = None
        self._check = None
        self._rhythm = None
        self._process = None
        self._format_in = None
        self._parent = None
        self._coerce = None


        self.rhythm = rhythm

        self.logger = init_logger(self)

        # self._wrap_process()

    @property
    def rhythm(self) -> TransformRhythm:
        return self._rhythm

    @rhythm.setter
    def rhythm(self, rhythm: TransformRhythm):
        if rhythm not in TransformRhythm:
            raise ValueError(f'rhythm must be one of TransformRhythm, got {rhythm}')
        self._rhythm = rhythm

    @property
    def format_in(self) -> dict:
        raise NotImplementedError('Every subclass of Transform must define format_in!')

    @format_in.setter
    def format_in(self, format_in: dict):
        raise NotImplementedError('Every subclass of Transform must define format_in!')

    @property
    def format_out(self) -> dict:
        raise NotImplementedError('Every subclass of Transform must define format_out!')

    @format_out.setter
    def format_out(self, format_out: dict):
        raise NotImplementedError('Every subclass of Transform must define format_out!')

    @property
    def parent(self) -> typing.Union['Transform', None]:
        """
        If this Transform is in a chain of transforms, the transform that precedes it

        Returns:
            :class:`.Transform`, ``None`` if no parent.
        """
        return self._parent

    @parent.setter
    def parent(self, parent):
        if not issubclass(type(parent), Transform):
            raise TypeError('parents must be subclasses of Transform')
        self._parent = parent


    def process(self, input):
        raise NotImplementedError('Every subclass of Transform must define its own process method!')

    def reset(self):
        """
        If a transformation is stateful, reset state.
        """
        raise Warning('reset method not explicitly overridden in transformation, doing nothing!')

    def check_compatible(self, child: 'Transform'):
        """
        Check that this Transformation's :attr:`.format_out` is compatible with another's :attr:`.format_in`

        .. todo::

            Check for types that can be automatically coerced into one another and set :attr:`_coercion` to appropriate function

        Args:
            child (:class:`Transform`): Transformation to check compatibility

        Returns:
            bool
        """

        ret = False

        if isinstance(child.format_in['type'], (list, tuple)):
            if self.format_out['type'] in child.format_in['type']:
                ret = True
        elif child.format_in['type'] == 'any':
            ret = True
        elif self.format_out['type'] == child.format_in['type']:
            ret = True


        # if child has a specific requirement of parent transform class, ensure
        parent_req = child.format_in.get('parent', False)
        if parent_req:
            if not isinstance(self, parent_req):
                ret = False

        return ret

        # if self.format_out['type'] in (int, np.int, )

    def __add__(self, other):
        """
        Add another Transformation in the chain to make a processing pipeline

        Args:
            other (:class:`Transformation`): The transformation to be chained
        """
        if not issubclass(type(other), Transform):
            raise RuntimeError('Can only add subclasses of Transform to other Transforms!')

        if self._child is None:
            # if we haven't been chained at all yet, claim the child
            # first check if it aligns

            #if not self.check_compatible(other):
            #    raise ValueError(f'Incompatible transformation formats: \nOutput: {self.format_out},\nInput: {other.format_in}')


            self._child = other
            self._child.parent = self

            # override our process method with one that calls recursively
            # back it up first
            self._process = self.process

            def new_process(self, input):
                return self._child.process(self._process(input))

            self.process = types.MethodType(new_process, self)

        else:
            # we already have a child,
            # add it to our child instead (potentially recursively)
            self._child = self._child + other

        return self





from time import sleep
from pathlib import Path
import json

import pytest
from pytestqt import qt_compat
from pytestqt.qt_compat import qt_api

from autopilot.core.terminal import Terminal
from autopilot import prefs


@pytest.fixture
def blank_pilot_db():
    pilot_db_fn = Path(prefs.get('PILOT_DB')).with_name('pilot_db_temp.json')
    with open(pilot_db_fn, 'w') as dbfile:
        json.dump({"testpilot_1": {
            "ip": "192.168.0.0",
            "subjects": ['subject_1', 'subject_2']
        }}, dbfile)
    prefs.set('PILOT_DB', str(pilot_db_fn))

    yield str(pilot_db_fn)

    pilot_db_fn.unlink()


@pytest.fixture
def spawn_terminal(qtbot, blank_pilot_db):
    prefs.clear()
    prefs.set('AGENT', 'TERMINAL')
    pilot_db_fn = blank_pilot_db
    prefs.set('PILOT_DB', pilot_db_fn)

    app = qt_api.QApplication.instance()
    app.setStyle('GTK+')
    terminal = Terminal()
    qtbot.addWidget(terminal)
    return app, terminal

def test_terminal_launch(qtbot, spawn_terminal):
    app, terminal = spawn_terminal

    sleep(5)

    assert terminal.isVisible()




"""Tests for generating sound stimuli.

This script runs tests that generate different sound stimuli and verifies
that they are initialized correctly.

Currently these only work if AUDIOSERVER is 'jack'. 'pyo' is not tested.
'docs' doesn't actually generate waveforms.

This doesn't require (or test) a running jackd or even a JackClient.
Instead, these tests short-circuit those dependencies by manually setting
FS and BLOCKSIZE in autopilot.stim.sound.jackclient.

A TODO is to test the JackClient itself.

Currently only the sound Noise is tested.

These tests cover multiple durations and amplitudes of mono and
multi-channel Noise, including some edges cases like very short durations
or zero amplitude.

The rest of this docstring addresses the workaround used to short-circuit
jackd and JackClient.

Here is the sequence of events that leads to FS and BLOCKSIZE.
* If an autopilot.core.pilot.Pilot is initialized:
**  autopilot.core.pilot.Pilot.__init__ checks prefs.AUDIOSERVER,
    and calls autopilot.core.pilot.Pilot.init_audio.
**  autopilot.core.pilot.Pilot.init_audio calls 
    autopilot.external.__init__.start_jackd.
**  autopilot.external.__init__.start_jackd takes the JACKDSTRING pref 
    and replaces the token '-rfs' in it with the FS pref. The jackd
    process is launched and stored in autopilot.external.JACKD_PROCESS.
    That process may fail or not, we continue anyway.
**  Next, autopilot.core.pilot.Pilot.init_audio instantiates an
    autopilot.stim.sound.jackclient.JackClient()
**  autopilot.stim.sound.jackclient.JackClient.__init__ 
    initalizes a jack.Client
**  autopilot.stim.sound.jackclient.JackClient.fs 
    is set to jack.Client.samplerate. Note that this is either the
    requested sample rate, or some default value from jack (not Autopilot)
    if the client did not actually succeed in booting.
**  autopilot.stim.sound.jackclient.FS (a global variable) is set to 
    autopilot.stim.sound.jackclient.JackClient.fs

* Later, a sound (e.g., Noise) is initialized.
**  autopilot.stim.sound.sounds.Noise.__init__ calls super().__init__,
**  which is autopilot.stim.sound.sounds.Jack_Sound.__init__
**  autopilot.stim.sound.sounds.Jack_Sound.__init__ 
    sets `self.fs` to jackclient.FS
**  autopilot.stim.sound.sounds.Noise.__init__ calls 
    autopilot.stim.sound.sounds.Noise.init_sound
**  autopilot.stim.sound.sounds.Noise.init_sound calls 
    autopilot.stim.sound.sounds.Jack_Sound.get_nsamples
**  autopilot.stim.sound.sounds.Jack_Sound.get_nsamples 
    inspects `self.fs`

To remove the dependence on jackd2 and JackClient, the entire first block 
of code can be circumvented by setting these:
autopilot.stim.sound.jackclient.FS
autopilot.stim.sound.jackclient.BLOCKSIZE
"""

import pytest
import numpy as np
import multiprocessing
import autopilot.prefs
import autopilot.external
import autopilot.stim.sound
from autopilot.stim.sound import jackclient, sounds


## Ensure we get the same random sound every time
np.random.seed(0)


## Specify needed params to circumvent init_audio
sample_rate = 192000
block_size = 1024
jackclient.FS = sample_rate
jackclient.BLOCKSIZE = block_size

# These are only used in Jack_Sound.__del__
# Setting them here to avoid warnings during garbage collection
# Why is there one global PLAY and STOP, rather than sound-specific?
jackclient.PLAY = multiprocessing.Event()
jackclient.STOP = multiprocessing.Event()


## Define tests
@pytest.mark.parametrize(
    "duration_ms,amplitude,check_duration_samples,check_n_chunks_expected",
    (
        (100, 0.1, 19200, 19),
        (100., 0.1, 19200, 19),
        ((block_size/sample_rate)*1000, 0.1, block_size, 1), # exactly one chunk
        (5, 0.1, 960, 1),
        (5., 0.1, 960, 1),
        (1/192000. - 1e-8, 0.1, 1, 1),
        (0, 0.1, 0, 0),
        ###### amplitude = 1
        (100, 1, 19200, 19),
        (100., 1, 19200, 19),
        ((block_size / sample_rate) * 1000, 1, block_size, 1),  # exactly one chunk
        (5, 1, 960, 1),
        (5., 1, 960, 1),
        (1 / 192000. - 1e-8, 1, 1, 1),
        (0, 1, 0, 0),
        ###### amplitude = 10.
        (100, 10., 19200, 19),
        (100., 10., 19200, 19),
        ((block_size / sample_rate) * 1000, 10., block_size, 1),  # exactly one chunk
        (5, 10., 960, 1),
        (5., 10., 960, 1),
        (1 / 192000. - 1e-8, 10., 1, 1),
        (0, 10., 0, 0)
    )
)
def test_init_noise(duration_ms, amplitude, 
    check_duration_samples, check_n_chunks_expected):
    """Initialize and check a mono (single-channel) noise.
    
    A mono `Noise` is initialized with specified duration and amplitude.
    The following things are checked:
    * The attributes should be correctly set
    * The `table` should be the right dtype and the right duration, 
      given the sampling rate
    * The chunks should be correct, given the block size. The last chunk
      should be zero-padded.
    * The waveform should not exceed amplitude anywhere
    * As long as the waveform is sufficiently long, it should exceed
      90% of the amplitude somewhere
    * Concatenating the chunks should generate a result equal to the
      table, albeit zero-padded to a multiple of the block size.
    * Specifying channel as None should give identical results to leaving
      it unspecified.
    
    Arguments
    ---------
    duration_ms : passed as `duration`
    amplitude : passed as `amplitude`
    check_duration_samples : int or None
        If not None, the length of the sounds `table` should be this
    check_n_chunks_expected : int or None
        If not None, the length of the sounds `chunks` should be this
    """
    ## Calculate how long the sound should be
    duration_samples = int(np.ceil(duration_ms / 1000. * sample_rate))
    n_chunks_expected = int(np.ceil(duration_samples / block_size))

    # Compare versus the requested checks
    if check_duration_samples is not None:
        assert check_duration_samples == duration_samples
    if check_n_chunks_expected is not None:
        assert check_n_chunks_expected == n_chunks_expected
    
    # Calculate number of padded zeros expected
    n_padded_zeros = n_chunks_expected * block_size - duration_samples
    assert n_padded_zeros >= 0
    assert n_padded_zeros < block_size

    # Test both when channel is explicitly set as None, and when it
    # is left unspecified
    for specify_channel_as_none in [True, False]:
        ## Init sound
        if specify_channel_as_none:
            noise = autopilot.stim.sound.sounds.Noise(
                duration=duration_ms, amplitude=amplitude, channel=None)
        else:
            noise = sounds.Noise(
                duration=duration_ms, amplitude=amplitude)

        
        ## Test attributes
        assert noise.channel is None
        assert noise.duration == float(duration_ms)
        assert noise.amplitude == float(amplitude)
        assert noise.initialized is True
        
        
        ## Test waveform
        # The table should be float32
        assert noise.table.dtype == np.float32
        
        # The table should be 1-dimensional with length duration_samples
        assert noise.table.shape == (duration_samples,)
        
        # The table should not exceed `amplitude` anywhere
        assert np.all(np.abs(noise.table) < amplitude)
        
        # The table itself should NOT be zero-padded
        # Vanishingly unlikely that any real sample is exactly zero
        assert np.all(noise.table != 0)

        # As long as we have enough samples, almost certainly
        # the max value should be >90% of amplitude.
        if duration_samples > 100:
            assert (np.abs(noise.table).max() > .9 * amplitude)

        # The chunks should each be length block_size
        assert len(noise.chunks) == n_chunks_expected
        assert np.all(np.array(list(map(len, noise.chunks))) == block_size)

        # The last chunk should be padded with zeros
        if n_padded_zeros > 0:
            assert np.all(noise.chunks[-1][-n_padded_zeros:] == 0)

        # Concatenate the chunks
        if duration_samples > 0:
            concatted = np.concatenate(noise.chunks)
        else:
            # Special case because np.concatenate([]) is an error
            concatted = np.array([], dtype=np.float32)
        
        # The concatenated chunks should be equal to the table
        assert len(concatted) == len(noise.table) + n_padded_zeros
        assert np.all(concatted[:len(noise.table)] == noise.table)

@pytest.mark.parametrize(
    "duration_ms,amplitude,channel,check_duration_samples,check_n_chunks_expected",
    (
        ########
        ######## channel = 0
        (100, 0.1, 0, 19200, 19),
        (100., 0.1, 0, 19200, 19),
        ((block_size/sample_rate)*1000, 0.1, 0, block_size, 1), # exactly one chunk
        (5, 0.1, 0, 960, 1),
        (5., 0.1, 0, 960, 1),
        (1/192000. - 1e-8, 0.1, 0, 1, 1),
        (0, 0.1, 0, 0, 0),
        ###### amplitude = 1
        (100, 1, 0, 19200, 19),
        (100., 1, 0, 19200, 19),
        ((block_size / sample_rate) * 1000, 1, 0, block_size, 1),  # exactly one chunk
        (5, 1, 0, 960, 1),
        (5., 1, 0, 960, 1),
        (1 / 192000. - 1e-8, 1, 0, 1, 1),
        (0, 1, 0, 0, 0),
        ###### amplitude = 10.
        (100, 10., 0, 19200, 19),
        (100., 10., 0, 19200, 19),
        ((block_size / sample_rate) * 1000, 10., 0, block_size, 1),  # exactly one chunk
        (5, 10., 0, 960, 1),
        (5., 10., 0, 960, 1),
        (1 / 192000. - 1e-8, 10., 0, 1, 1),
        (0, 10., 0, 0, 0),
        #######
        ####### channel = 1
        (100, 0.1, 1, 19200, 19),
        (100., 0.1, 1, 19200, 19),
        ((block_size/sample_rate)*1000, 0.1, 1, block_size, 1), # exactly one chunk
        (5, 0.1, 1, 960, 1),
        (5., 0.1, 1, 960, 1),
        (1/192000. - 1e-8, 0.1, 1, 1, 1),
        (0, 0.1, 1, 0, 0),
        ###### amplitude = 1
        (100, 1, 1, 19200, 19),
        (100., 1, 1, 19200, 19),
        ((block_size / sample_rate) * 1000, 1, 1, block_size, 1),  # exactly one chunk
        (5, 1, 1, 960, 1),
        (5., 1, 1, 960, 1),
        (1 / 192000. - 1e-8, 1, 1, 1, 1),
        (0, 1, 1, 0, 0),
        ###### amplitude = 10.
        (100, 10., 1, 19200, 19),
        (100., 10., 1, 19200, 19),
        ((block_size / sample_rate) * 1000, 10., 1, block_size, 1),  # exactly one chunk
        (5, 10., 1, 960, 1),
        (5., 10., 1, 960, 1),
        (1 / 192000. - 1e-8, 10., 1, 1, 1),
        (0, 10., 1, 0, 0)
    )
)
def test_init_multichannel_noise(duration_ms, amplitude, channel, 
    check_duration_samples, check_n_chunks_expected):
    """Initialize and check a multi-channel noise.

    A multi-channel `Noise` is initialized with specified duration, amplitude,
    and channel. The following things are checked:
    * The attributes should be correctly set
    * The `table` should be the right dtype and the right duration, 
      given the sampling rate
    * The chunks should be correct, given the block size. The last chunk
      should be zero-padded.
    * The column `channel` should contain non-zero data and all other
      columns should contain zero data.
    * The waveform should not exceed amplitude anywhere
    * As long as the waveform is sufficiently long, it should exceed
      90% of the amplitude somewhere
    * Concatenating the chunks should generate a result equal to the
    
    Arguments
    ---------
    duration_ms : passed to `Noise` as `duration`
    amplitude : passed to `Noise` as `amplitude`
    channel : passed to `Noise` as `channel`
    check_duration_samples : int or None
        If not None, the length of the sounds `table` should be this
    check_n_chunks_expected : int or None
        If not None, the length of the sounds `chunks` should be this
    """
    ## Calculate how long the sound should be
    duration_samples = int(np.ceil(duration_ms / 1000. * sample_rate))
    n_chunks_expected = int(np.ceil(duration_samples / block_size))

    # Compare versus the requested checks
    if check_duration_samples is not None:
        assert check_duration_samples == duration_samples
    if check_n_chunks_expected is not None:
        assert check_n_chunks_expected == n_chunks_expected
    
    # Calculate number of padded zeros expected
    n_padded_zeros = n_chunks_expected * block_size - duration_samples
    assert n_padded_zeros >= 0
    assert n_padded_zeros < block_size

    
    ## Init sound
    noise = sounds.Noise(
        duration=duration_ms, amplitude=amplitude, channel=channel)


    ## Test attributes
    assert noise.channel == channel
    assert noise.duration == float(duration_ms)
    assert noise.amplitude == float(amplitude)
    assert noise.initialized is True
    
    
    ## Tests
    # The table should be float32
    assert noise.table.dtype == np.float32
    
    # The table should be 2-dimensional with length duration_samples
    assert noise.table.shape == (duration_samples, 2)

    # The table should not exceed `amplitude` anywhere
    assert (np.abs(noise.table) < amplitude).all()

    # Check each column of the table
    for n_col in range(noise.table.shape[1]):
        # Only the `channel` column should contain data
        if n_col == channel:
            # The table itself should NOT be zero-padded
            # Vanishingly unlikely that any real sample is exactly zero
            assert (noise.table[:, n_col] != 0).all()
            
            # As long as we have enough samples, almost certainly
            # the max value should be >90% of amplitude.
            if duration_samples > 100:
                assert (np.abs(noise.table[:, n_col]).max() > .9 * amplitude)
        else:
            # Other channels should be all zero
            assert (noise.table[:, n_col] == 0).all()

    # The chunks should each be shape (block_size, 2)
    assert len(noise.chunks) == n_chunks_expected
    for chunk in noise.chunks:
        assert chunk.shape == (block_size, 2)

    # The last chunk should be padded with zeros
    if n_padded_zeros > 0:
        assert np.all(noise.chunks[-1][-n_padded_zeros:, :] == 0)

    # Concatenate the chunks
    if duration_samples > 0:
        concatted = np.concatenate(noise.chunks)
    else:
        # Special case because np.concatenate([]) is an error
        # Make sure shape is (0, 2)
        concatted = np.array([[], []], dtype=np.float32).T
    
    # The concatenated chunks should be equal to the table
    assert concatted.shape == (len(noise.table) + n_padded_zeros, 2)
    assert (concatted[:len(noise.table)] == noise.table).all()

def test_unpadded_gap():
    """
    A gap in a continous sound should not be padded (had its last chunk filled with zeros).
    """

    # make a sound that is 1.5x the duration of a chunk
    nsamples = round(block_size*1.5)
    duration = round((nsamples/sample_rate)*1000)
    gap = sounds.Gap(duration)

    assert len(gap.chunks) == 2
    assert len(gap.chunks[1]) == nsamples % block_size
    assert len(gap.chunks[1]) != block_size



import altair as alt
from sklearn.linear_model import LogisticRegression
from autopilot import prefs
from autopilot.core.subject import Subject
import numpy as np
from autopilot.utils.common import coerce_discrete
import pandas as pd
import pdb


def calc_psychometric(data, var_x, var_y='response'):
    """
    Calculate a psychometric curve (logistic regression of var_y on var_x)

    Args:
        data (:class:`pandas.DataFrame`): Subject data
        var_x (str): name of column to use as the discriminand
        var_y (str): name of the column for the response, usually 'response'

    Returns:
        params (tup): parameters for logistic function
    """
    data = data[[var_x, var_y]].copy()

    try:
        if not all(data[var_y].str.isnumeric()):
            data = coerce_discrete(data, var_y)
    except AttributeError:
        # if w can't use the .str accessor, then it's probs a number
        pass



    log_regress = LogisticRegression()
    try:
        log_regress.fit(data[var_x], data[var_y])
    except ValueError:
        # need to have X data be 2d
        log_regress.fit(data.as_matrix(columns=[var_x]), data[var_y])

    return log_regress


def plot_psychometric(subject_protocols):
    """
    Plot psychometric curves for selected subjects, steps, and variables

    Typically called by :meth:`.Terminal.plot_psychometric`.

    Args:
        subject_protocols (list): A list of tuples, each with

            * subject_id (str)
            * step_name (str)
            * variable (str)

    Returns:
        :class:`altair.Chart`
    """

    for subject, step, var, n_trials in subject_protocols:
        # load subject dataframe and subset
        asub = Subject(subject)
        sub_df = asub.get_trial_data(step)

        if n_trials>0:
            sub_df = sub_df[-n_trials:]

        # pdb.set_trace()

        sub_df = coerce_discrete(sub_df, 'response')
        logit = calc_psychometric(sub_df, var)

        # generate points to plot regression fit
        logit_x = np.linspace(sub_df[var].min(), sub_df[var].max(), 100).reshape(-1,1)
        logit_y = logit.predict_proba(logit_x)

        # pdb.set_trace()
        this_logit_df = pd.DataFrame({'x': logit_x[:,0], 'y':logit_y[:,1]})
        this_logit_df['subject'] = subject
        this_logit_df['type'] = 'log_regression'

        try:
            logit_df = pd.concat([logit_df, this_logit_df])
        except NameError:
            logit_df = this_logit_df

        # and also mean accuracy by var
        sub_df_sum = sub_df.groupby(var).mean().reset_index()[[var, 'response']]
        sub_df_sum['subject'] = subject
        sub_df_sum['type'] = 'responses'
        sub_df_sum.rename(columns={var:'x', 'response':'y'}, inplace=True)

        try:
            sum_df = pd.concat([sum_df, sub_df_sum])
        except NameError:
            sum_df = sub_df_sum

    # combine dataframes
    combo_df = pd.concat([sum_df, logit_df])

    acc_points = alt.Chart().encode(
        alt.X('x:Q',scale=alt.Scale(type='sqrt'), title=var),
        y=alt.Y('y:Q',title="mean response")
    ).transform_filter(
        alt.FieldEqualPredicate('responses', 'type')
    ).mark_point()

    log_curves = alt.Chart().encode(
        alt.X('x:Q',scale=alt.Scale(type='sqrt'), title=var),
        y=alt.Y('y:Q',title="mean response")
    ).transform_filter(
        alt.FieldEqualPredicate('log_regression', 'type')
    ).mark_line()

    combo = alt.layer(acc_points + log_curves, data=combo_df).facet(row='subject:N')
    #combo.sav
    return combo

    #combo.serve()

    # acc_points.serve()











import numpy as np
from autopilot.transform.transforms import Transform

class Add(Transform):
    def __init__(self, value=0, *args, **kwargs):
        super(Add, self).__init__(*args, **kwargs)

        self.value = value

    def process(self, input):
        return input + self.value


"""
Go/no-go task, demo for NCB
"""



import datetime
import itertools
import tables
import threading
from random import random

import autopilot.hardware.gpio
from autopilot.tasks import Task
from autopilot.stim.visual.visuals import Grating
from collections import OrderedDict as odict
from autopilot.networking import Net_Node

from autopilot import prefs
TASK = 'GoNoGo'

class GoNoGo(Task):
    """
    A Visual Go/No-Go task using a :class:`.Pilot` and a :class:`.Wheel_Child`.

    .. note::

        This task was written as a proof-of-concept for the Autopilot manuscript,
        and is thus underdeveloped and underdocumented, submit and issue if you would like to
        use it yourself :)

    """

    STAGE_NAMES = ["request", "discrim", "reinforcement"]

    # Class attributes

    # List of needed params, returned data and data format.
    # Params are [name]={'tag': Human Readable Tag, 'type': 'int', 'float', 'bool', etc.}
    PARAMS = odict()
    PARAMS['reward'] = {'tag': 'Reward Duration (ms)',
                        'type': 'int'}
    PARAMS['timeout']        = {'tag':'Delay Timeout (ms)',
                                'type':'int'}
    PARAMS['stim'] = {'tag':  'Visuals',
                      'type': 'visuals'}

    # Set plot params, which data should be plotted, its default shape, etc.
    PLOT = {
        'data': {
            'y': 'shaded',
            'target': 'point',
            'response': 'segment'
        },
        'continuous': True
    }

    # PyTables Data descriptor
    # for numpy data types see http://docs.scipy.org/doc/numpy/reference/arrays.dtypes.html#arrays-dtypes-constructing
    class TrialData(tables.IsDescription):
        # This class allows the Subject object to make a data table with the correct data types. You must update it for any new data you'd like to store
        trial_num = tables.Int32Col()
        target = tables.BoolCol()
        response = tables.StringCol(1)
        correct  = tables.Int32Col()
        RQ_timestamp = tables.StringCol(26)
        DC_timestamp = tables.StringCol(26)
        shift = tables.Float32Col()
        angle = tables.Float32Col()
        delay = tables.Float32Col()

    # class ContinuousData(tables.IsDescription):
    #     x = tables.Float64Col()
    #     y = tables.Float64Col()
    #     t = tables.Float64Col()

    HARDWARE = {
        'POKES': {
            'C': autopilot.hardware.gpio.Digital_In,
        },
        'LEDS': {
            'C': autopilot.hardware.gpio.LED_RGB,
        },
        'PORTS': {
            'C': autopilot.hardware.gpio.Solenoid,
        },
        'FLAGS': {
            'F': autopilot.hardware.gpio.Digital_Out
        }
    }

    CHILDREN = {
        'WHEEL': {
            'task_type': "Wheel Child",
        }
    }

    def __init__(self, stim=None, reward = 50, timeout = 1000, stage_block = None,**kwargs):
        super(GoNoGo, self).__init__()

        self.stage_block = stage_block
        self.trial_counter = itertools.count()

        self.punish_dur = 500.0

        self.reward = reward
        self.timeout = timeout

        self.init_hardware()
        self.set_reward(self.reward)

        self.node = Net_Node(id="T_{}".format(prefs.get('NAME')),
                             upstream=prefs.get('NAME'),
                             port=prefs.get('MSGPORT'),
                             listens={},
                             instance=True)

        # get our child started
        self.subject = kwargs['subject']
        value = {
            'child': {'parent': prefs.get('NAME'), 'subject': kwargs['subject']},
            'task_type': 'Wheel Child',
            'subject': kwargs['subject']
        }

        self.node.send(to=prefs.get('NAME'), key='CHILD', value=value)

        # hardcoding stimulus for testing
        self.stim = Grating(angle=0, freq=(4,0), rate=1, size=(1,1), debug=True)

        self.stages = itertools.cycle([self.request, self.discrim, self.reinforce])


    def request(self):
        # wait for the subject to hold the wheel still
        # Set the event lock
        self.stage_block.clear()
        # wait on any ongoing punishment stimulus
        self.punish_block.wait()

        # Reset all the variables that need to be
        #for v in self.resetting_variables:
        #    v = None

        # reset triggers if there are any left
        self.triggers = {}

        # calculate orientation change
        # half the time, don't change, otherwise, do change
        if random() < 0.5:
            self.shift = 0
            self.target = False
        else:
            self.shift = random()*180.0
            self.target = True



        # Set sound trigger and LEDs
        # We make two triggers to play the sound and change the light color
        change_to_blue = lambda: self.hardware['LEDS']['C'].set_color([0, 0, 255])

        # set triggers
        self.triggers['F'] = [change_to_blue, lambda: self.stim.play('shift', self.shift )]

        # set to green in the meantime
        self.set_leds({'C': [0, 255, 0]})

        # tell our wheel to start measuring
        self.node.send(to=[prefs.get('NAME'), prefs.get('CHILDID'), 'wheel_0'],
                       key="MEASURE",
                       value={'mode':'steady',
                              'thresh':100})

        self.current_trial = next(self.trial_counter)
        data = {
            'target': self.target,
            'shift': self.shift,
            'trial_num': self.current_trial
        }

        self.current_stage = 0
        return data

    def discrim(self):
        self.stage_block.clear()

        # if the subject licks on a good trial, reward.
        # set a trigger to respond false if delay time elapses
        if self.target:
            self.triggers['C'] = [lambda: self.respond(True), self.hardware['PORTS']['C'].open]
            self.triggers['T'] = [lambda: self.respond(False), self.punish]
        # otherwise punish
        else:
            self.triggers['C'] = [lambda: self.respond(True), self.punish]
            self.triggers['T'] = [lambda: self.respond(False), self.hardware['PORTS']['C'].open]

        # the stimulus has just started playing, wait a bit and then shift it (if we're gonna
        # choose a random delay
        delay = 0.0
        if self.shift != 0:
            delay = (random()*3000.0)+1000.0
            self.delayed_set(delay, 'shift', self.shift)

        self.timer = threading.Timer(5.0, self.handle_trigger, args=('T', True, None)).start()

        data = {
            'delay': delay,
            'RQ_timestamp': datetime.datetime.now().isoformat(),
            'trial_num': self.current_trial
        }


        return data





    def reinforce(self):
        # don't clear stage block here to quickly move on
        if self.response == self.target:
            self.correct = 1
        else:
            self.correct = 0

        self.set_leds({'C': [0, 0, 0]})
        # stop timer if it's still going
        try:
            self.timer.cancel()
        except AttributeError:
            pass
        self.timer = None

        data = {
            'DC_timestamp': datetime.datetime.now().isoformat(),
            'response': self.response,
            'correct': self.correct,
            'trial_num': self.current_trial,
            'TRIAL_END': True
        }
        return data

    def respond(self, response):
        self.response = response

    def delayed_set(self, delay, attr, val):
        threading.Timer(float(delay)/1000.0, self._delayed_shift, args=(attr,val)).start()

    def _delayed_shift(self, attr, val):
        self.stim.q.put(('shift', val))

    def punish(self):
        """
        Flash lights, play punishment sound if set
        """
        # TODO: If we're not in the last stage (eg. we were timed out after stim presentation), reset stages
        self.punish_block.clear()

        if self.punish_stim:
            self.stim_manager.play_punishment()

        # self.set_leds()
        self.flash_leds()
        threading.Timer(self.punish_dur / 1000., self.punish_block.set).start()




"""
Utility functions for handling plugins, eg. importing, downloading, listing, confirming, etc.
"""
import sys
import inspect
import importlib
from typing import Union, Optional
from pathlib import Path
from multiprocessing import Value
from ctypes import c_bool
from collections import OrderedDict as odict

import requests

from autopilot.core.loggers import init_logger
from autopilot import prefs
from autopilot.utils import wiki

_IMPORTED = Value(c_bool, False)

def import_plugins(plugin_dir:Optional[Path]=None) -> dict:
    """
    Import all plugins in the plugin (or supplied) directory.

    There is no specific form for a plugin at the moment, so
    this function will recursively import all modules and packages
    within the directory.

    Plugins can then be accessed by the :func:`~.utils.registry.get` registry
    functions.

    Args:
        plugin_dir (None, :class:`pathlib.Path`): Directory to import. if
            ``None`` (default), use ``prefs.get('PLUGINDIR')``.

    Returns:
        dict: of imported objects with form {"class_name": class_object}
    """
    logger = init_logger(module_name='plugins')

    if plugin_dir is not None:
        plugin_dir = Path(plugin_dir)
    else:
        plugin_dir = Path(prefs.get('PLUGINDIR'))

    if not plugin_dir.exists():
        logger.exception(f"Plugin directory {plugin_dir} does not exist!")
        return {}

    plugins = {}

    # recursively list python files in plugin directory
    plugin_files = Path(plugin_dir).glob('**/*.py')
    for pfile in plugin_files:
        # prepend autopilot.plugins to avoid namespace clashes and make them uniformly moduled
        module_name = 'autopilot.plugins.' + inspect.getmodulename(str(pfile))

        # import module
        try:
            if module_name in sys.modules:
                mod = sys.modules[module_name]
            else:
                mod_spec = importlib.util.spec_from_file_location(module_name, pfile)
                mod = importlib.util.module_from_spec(mod_spec)
                mod_spec.loader.exec_module(mod)
                sys.modules[module_name] = mod
        except Exception as e:
            logger.exception(f'plugin file {str(pfile)} could not be imported, got exception: {e}')
            continue

        # get the imported modules and check if they're one of ours
        # filter for classes and then for classes that are declared in the plugin
        # module itself
        members = inspect.getmembers(mod, inspect.isclass)
        members = [member for member in members if member[1].__module__ == module_name]

        # store in the plugin dictionary to be returned,
        for member in members:
            # if a plugin object with the same name was already found, warn the user and don't overwrite
            if member[0] in plugins.keys():
                logger.warning(f'Conflicting plugin object names, importing both but only returning first found: {member[0]}\nfound first in: {plugins[member[0]]}\nand now in: {member[1].__module__}')
                continue
            plugins[member[0]] = member[1]

    globals()['_IMPORTED'].value = True

    return plugins



def unload_plugins():
    """
    Un-import imported plugins (mostly for testing purposes)
    """
    mods = [mod for mod in sys.modules if 'autopilot.plugins' in mod]
    for mod in mods:
        del sys.modules[mod]
    _IMPORTED.value = False


def list_wiki_plugins():
    """
    List plugins available on the wiki using :func:`.utils.wiki.ask`

    Returns:
        dict: {'plugin_name': {'plugin_prop':'prop_value',...}
    """

    plugins = wiki.ask(filters="[[Category:Autopilot Plugin]]",
             properties=[
                 'Created By',
                 'Has Description',
                 'For Autopilot Version',
                 'Has Git Repository',
                 'Has Contributor',
                 'Is Autopilot Plugin Type',
                 'Controls Hardware',
                 'Has DOI'
             ])

    ordered_plugins = []
    # reorder fields in plugins with odicts
    for plugin in plugins:
        oplugin = odict()
        for field in ('name','Has Description',
                      'Created By', 'Has Contributor',
                      'url','Has Git Repository',
                      'Has DOI',
                      'For Autopilot Version',
                      'Is Autopilot Plugin Type',
                      'Controls Hardware'):
            oplugin[field] = plugin[field]

        ordered_plugins.append(oplugin)

    return plugins



